package edu.toronto.cs.algebra;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.IntIterator.*;
import java.util.*;
import java.io.*;

import edu.toronto.cs.algebra.OrderedSet.SetElement;

/****
 **** A finite lattice
 ****/
public class FiniteLattice implements Algebra
{

  public static final int MEET = 0;
  public static final int JOIN = 1;
  public static final int NEG = 2;
  public static final int BELOW = 3;
  public static final int ABOVE = 4;
  public static final int EQUAL = 5;

  public static final int COMPOSE = 6;
  public static final int COMPOSE3 = 7;
  
  // -- this is an ordered set that actually computes the meet and join tables
  OrderedSet oSet;
  
  // -- our lattice has negation 
  int[] negationTable;
  
  // -- an image file for this lattice
  File imageFile = null;

  public FiniteLattice ()
  {
  }
  

  public FiniteLattice (FiniteLattice lat1, FiniteLattice lat2)
  {
    oSet = new OrderedSet (lat1.getOrderedSet (),
			   lat2.getOrderedSet ());
    buildNegationTable (lat1, lat2);
    
  }
  

  public FiniteLattice (Collection elements, Collection aboveRelation, 
			Collection negRelation)
  {
    oSet = new OrderedSet (elements, aboveRelation);
    buildNegationTable (negRelation);
  }

  public void setImageFile (File v)
  {
    imageFile = v;
  }
  public File getImageFile ()
  {
    return imageFile;
  }
  

  OrderedSet getOrderedSet ()
  {
    return oSet;
  }
  
  public int[] getNegTable ()
  {
    return negationTable;
  }
  
  public int[][] getMeetTable ()
  {
    return oSet.getMeetTable ();
  }
  
  public int[][] getJoinTable ()
  {
    return oSet.getJoinTable ();
  }
  
  void buildNegationTable (FiniteLattice lat1, FiniteLattice lat2)
  {
    negationTable = new int [oSet.size ()];
    
    SetElement[] elements1 = lat1.getOrderedSet ().getElements ();
    SetElement[] elements2 = lat2.getOrderedSet ().getElements ();

    for (int i = 0; i < elements1.length; i++)
      {
	for (int j = 0; j < elements2.length; j++)
	  {
	    int elemKey =
	      getElementIdxByKey (elements1 [i].getData ().toString () + 
				  elements2 [j].getData ().toString ());

	    int neg1 = lat1.unOp (elements1 [i].getOrd (), FiniteLattice.NEG);
	    int neg2 = lat2.unOp (elements2 [j].getOrd (), FiniteLattice.NEG);
	    
	    int negKey =
	      getElementIdxByKey (lat1.getElementKeyByIdx (neg1).toString () +
				  lat2.getElementKeyByIdx (neg2).toString ());
	    negationTable [elemKey] = negKey;
	    
	  }
      }
    
    
  }
  
  void buildNegationTable (Collection negRelation)
  {
    negationTable = new int [oSet.size ()];
    
    for (Iterator it = negRelation.iterator (); it.hasNext ();)
      {
	Object[] neg = (Object[])it.next ();

	// -- ~a = b
	Object a = neg [0];
	Object b = neg [1];
	negationTable [oSet.getElementByKey (a).getOrd ()] 
	  = oSet.getElementByKey (b).getOrd ();
      }
    
  }
  

  public int binOp (int l, int r, int opcode) throws NotAnOperationException
  {
    switch (opcode)
      {
      case MEET:
	return oSet.meet (l, r);
      case JOIN:
	return oSet.join (l, r);
      case BELOW:
	return leq (l, r) ? idtimesIdx () : idplusIdx ();
      case ABOVE:
	return geq (l, r) ? idtimesIdx () : idplusIdx ();
      case EQUAL:
	return eq (l, r) ? idtimesIdx () : idplusIdx ();
      case COMPOSE:
	return compose (l, r);
      case COMPOSE3:
	return compose3 (l, r);
      default:
	throw new NotAnOperationException ("Unsupported opcode: " + opcode);
      }
  }

  public int compose (int l, int r)
  {
    int tt = getElementIdxByKey ("TT");
    int tf = getElementIdxByKey ("TF");
    int ft = getElementIdxByKey ("FT");
    int ff = getElementIdxByKey ("FF");
    
    if (l == r) return l;
    if (l == tt && r == ff) return tf;
    return ft;
  }
  public int compose3 (int l, int r)
  {
    if (l != r)
      {
	return getElementIdxByKey ("M");
      }
    else return l;
  }
  

  public int neg (int l)
  {
    return unOp (l, NEG);
  }
  
  public int unOp (int l, int opcode) throws NotAnOperationException
  {
    switch (opcode)
      {
      case NEG:
	return negationTable [l];
      default:
	throw new NotAnOperationException ("Unsupported opcode: " + opcode);
      }
  }
  
  public int plus (int l, int r) throws NotInCarrierSetException
  {
    return oSet.join (l, r);
  }
  
  public int idplusIdx ()
  {
    return oSet.getBottomIdx ();
  }
  
  
  public int times (int l, int r) throws NotInCarrierSetException
  {
    return oSet.meet (l, r);
  }
  
  public int idtimesIdx ()
  {
    return oSet.getTopIdx ();
  }
  
  public int size ()
  {
    return oSet.size ();
  }
  
  public IntIterator carrierSet ()
  {
    return oSet.carrierSet ();
  }
  
  // -- eq, geq, leq
  public boolean eq (int a, int b)
  {
    return a == b;
  }
  
  // -- computes a under b
  public boolean leq (int a, int b)
  {
    // -- a <= b iff a \/ b = b
    return oSet.join (a, b) == b;
  }
  public boolean geq (int a, int b)
  {
    // -- a >= b iff b <= a
    return leq (b, a);
  }
  public boolean isComparable (int a, int b)
  {
    // -- a is comparable to b iff a <= b or b <= a
    return leq (a, b) || geq (a, b);
  }
  
    
  public int getElementIdxByKey (Object key)
  {
    OrderedSet.SetElement element = oSet.getElementByKey (key);
    return element != null ? element.getOrd () : Algebra.NO_VALUE;
  }
  
  public Object getElementKeyByIdx (int idx)
  {
    return oSet.getElement (idx).getData ();
  }

  /** This method will take a BitSet representing lattice values and
      clear all those bits which are redundant under join, ie. are under
      other bits.  
      @param Interested the BitSet representing the values you're 
      interested in (will not be touched.)  
      @return BitSet marking irredundant elements.  
      @author CDTW 
  */
  public BitSet getIrredundantJoin(BitSet interested)
  {
    // XXX CDTW NOTE: this is dependant on the index into the lattice
    // yielding topological order when iterated through; it is thus
    // dependant on our lattice representation.
    BitSet irredundant = (BitSet)interested.clone ();
    // FiniteLattice lat=(FiniteLattice)lattice;
    for (int i = 0; i < irredundant.length (); i++)
      {
	// -- skip empty bits
	if (!irredundant.get(i)) continue;
	  
	for (int j = i + 1; j < irredundant.length (); j++)
	  if (leq (j, i))
	    irredundant.clear(j);
      }
      
    return irredundant;
  }
  
  /***
   *** Given a set of lattice values returns 'min' of it 
   *** i.e. the lowest elements only
   *** Assumes topological sort of lattice values
   ***/
  public BitSet getIrredundantMeet (BitSet interested)
  {
    BitSet irredundant = (BitSet)interested.clone ();
    for (int i = 0; i < irredundant.length (); i++)
      {
	if (!irredundant.get (i)) continue;
	
	for (int j = i + 1; j < irredundant.length (); j++)
	  if (irredundant.get (j) && leq (j, i)) 
	    {
	      irredundant.clear (i);
	      break;
	    }
      }
    return irredundant;
  }
  
  

  // -- debug stuff
  public void dumpTables (PrintWriter out)
  {
    out.println ("Dumping negation table");
    dumpNegation (out);
    oSet.dumpTables (out);
  }
  
  void dumpNegation (PrintWriter out)
  {
    out.println ("Negation: ");
    for (int i = 0; i < negationTable.length; i++)
      out.println ("~" + oSet.getElement (i) + " = " + 
		   oSet.getElement (negationTable [i]));
  }

  
  public void dumpValueMap (PrintWriter out)
  {
    oSet.dumpValueMap (out);
  }


  int[] joinIrr = null;
  
  // -- get all join irreducibles
  public int[] getJoinIrreducibles ()
  {
    if (joinIrr == null) 
      joinIrr = computeJoinIrr ();
    return joinIrr;
  }
  
  // -- actually computes join irreducibles based on the join table
  int[] computeJoinIrr ()
  {
    // XXX -- This is very ugly and no thought was given to how to write it
    // XXX -- please rewrite if you have a better idea
    boolean[] irreducibles = new boolean [oSet.size ()];

    // -- first we assume that all elements are irreducibles
    Arrays.fill (irreducibles, true);

    // -- \bot is not a join irreducible by convention
    irreducibles [oSet.getBottomIdx ()] = false;

    // -- loop through all of the elements
    for (IntIterator it1 = oSet.carrierSet (); it1.hasNext ();)
      {
	int i = it1.nextInt ();
	for (IntIterator it2 = oSet.carrierSet (); it2.hasNext ();)
	  {
	    int j = it2.nextInt ();
	    
	    int k = oSet.join (i, j);
	    
	    // -- k = i \/ j , if k is join irreducible then 
	    // -- k = i or k = j, if this is not the case
	    // -- it is not a join irreducible
	    if (irreducibles [k] &&
		!eq (i, k) &&
		!eq (j, k))
	      irreducibles [k] = false;
	  }
      }

    // -- count how many join irreducibles we have
    int count = 0;
    for (int i = 0; i < irreducibles.length; i++)
      if (irreducibles [i]) count++;
    
    // -- create an array to hold the result
    int[] result = new int [count];
    
    // -- fill it in
    int idx = 0;
    for (int i = 0; i < irreducibles.length; i++)
      if (irreducibles [i])
	result [idx++] = i;
    
    return result;
  }
  

  // -- given a lattice value returns it's join decomposition
  public int[] joinDecomposition (int value)
  {
    // -- we need to do it twice, first to find out how many 
    // -- elements we have, then to get the elements
    int[] joinIrr = getJoinIrreducibles ();
    
    int count = 0;
    for (int i = 0; i < joinIrr.length; i++)
      if (leq (joinIrr [i], value)) count++;
    
    int[] result = new int [count];    
    int idx = 0;
    for (int i = 0; i < joinIrr.length; i++)
      if (leq (joinIrr [i], value)) result [idx++] = joinIrr [i];

    return result;
  }

  public int[] getUpSet (int value)
  {
    // XXX potentially we will have a lookup table
    return computeUpSet (value);
  }
  
  public int[] computeUpSet (int value)
  {
    // -- given a value returns a set of all elements geq then that value
    int[] up = new int [oSet.size ()];
    
    int idx = 0;
    for (IntIterator it = oSet.carrierSet (); it.hasNext ();)
      {
	int x = it.nextInt ();
	if (geq (x, value))
	  up [idx++] = x;
	
      }
    
    int[] result = new int [idx];
    System.arraycopy (up, 0, result, 0, idx);
    return result;
  }


  public static void main (String[] args) throws Exception
  {
    
    FiniteLattice lat1 = LatticeFactory.parse (args [0]);
    FiniteLattice lat2 = LatticeFactory.parse (args [1]);
    PrintWriter out = new PrintWriter (System.out);
    out.println ("Dumping tables");
    lat1.dumpTables (out);
    FiniteLattice product = new FiniteLattice (lat1, lat2);
    product = new FiniteLattice (product, product);
    product.dumpTables (out);
    out.flush ();
    
  }
  
  
}



