package edu.toronto.cs.algebra;

import java.util.*;
import java.io.PrintWriter;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.IntIterator.*;

/****
 **** Implementation of an OrderedSet that sorts its elements in topological
 **** order and builds a join table following the algorithms in 
 **** Free Lattices by Ralph Freese
 **** author: Arie Gurfinkel <arie@cs.toronto.edu>

/****
 **** Implements all kinds of (partially) ordered set operations
 **** namely: topological sorting and computing a join table
 ****/
public class OrderedSet
{

  // -- map that maps user idea of an element in the ordered set to 
  // -- its internal representation
  Map dataToElements;
  
  SetElement[] elements;
  Map upperCovers;
  Map lowerCovers;
  
  // -- the join table
  int[][] joinTable;
  int[][] meetTable;


  /*** Construct an ordered set which is a product of set1 and set2 **/
  public OrderedSet (OrderedSet set1, OrderedSet set2)
  {
    Collection elements = new ArrayList ();
    Collection aboveRelation = new ArrayList ();

    constrcutProduct (set1, set2, elements, aboveRelation);
    
    init (elements, aboveRelation);
    
  }
  
  
  /****
   **** Constructs an ordered set given a collection of elements and 
   **** an above relation (given a set of pairs) */
  public OrderedSet (Collection _elements, Collection aboveRelation)
  {
    init (_elements, aboveRelation);
  }

  void init (Collection _elements, Collection aboveRelation)
  {
    upperCovers = new HashMap ();
    lowerCovers = new HashMap ();
    // -- map the elements first
    mapElements (_elements);
    // -- construct upper covers
    constructUpperCovers (aboveRelation);
    constructLowerCovers (aboveRelation);
    
    elements = addWithOrder (dataToElements.values ());

    // -- reverse topological sort
    elements = sortElements (dataToElements.values ());

    // -- build join table
    buildJoinTable ();
    buildMeetTable ();
    
  }

  public int[][] getJoinTable ()
  {
    return joinTable;
  }
  public int[][] getMeetTable ()
  {
    return meetTable;
  }
  

  // -- tester method
  public static void main (String[] args)
  {
    List elem = new ArrayList ();
    elem.add ("T");
    elem.add ("S");
    elem.add ("N");
    elem.add ("C");
    elem.add ("U");
    elem.add ("F");
    
    List above = new ArrayList ();
    above.add (tuple ("T", "S"));
    above.add (tuple ("S", "N"));
    above.add (tuple ("S", "C"));
    above.add (tuple ("N", "U"));
    above.add (tuple ("C", "U"));
    above.add (tuple ("U", "F"));
    
    OrderedSet oSet = new OrderedSet (elem, above);
    
    printJoinTable (oSet);
    printMeetTable (oSet);
    
  }
  
  static Object[] tuple (Object o1, Object o2)
  {
    Object[] t = new Object [2];
    t [0] = o1;
    t [1] = o2;
    return t;
  }
  
  static void printJoinTable (OrderedSet oSet)
  {
    int[][] table = oSet.joinTable;
    printTable (oSet, table);
  }

  static void printMeetTable (OrderedSet oSet)
  {
    printTable (oSet, oSet.meetTable);
  }
  
  static void printTable (OrderedSet oSet, int[][] table)
  {
    
    System.out.println ("1,1 is " + oSet.getElement (table [1][1]));
    for (int i = 0; i < table.length; i++)
      {
	for (int j = 0; j < table [i].length; j++)
	  System.out.print (oSet.getElement (table [i][j]) + " ");
	System.out.println ();
      }
    
  }


  public void dumpTables (PrintWriter out)
  {
    out.println ("Join table: ");
    dumpTable (out, joinTable);
    out.println ("Meet table: ");
    dumpTable (out, meetTable);
  }
  
  public void dumpValueMap (PrintWriter out)
  {
    for (int i = 0; i < elements.length; i++)
      out.println (getElement(i).getData ()+ " = " + i);
  }
  
  void dumpTable (PrintWriter out, int[][] table)
  {
    for (int i = 0; i < table.length; i++)
      {
	for (int j = 0; j < table [i].length; j++)
	  out.print (getElement (table [i][j]) + " ");
	out.println ();
      }
  }
  
  public int size ()
  {
    return dataToElements.size ();
  }
  
  public IntIterator carrierSet ()
  {
    return new RangeIterator (0, size ());
  }
  

  /*** Create an array of SetElement, populate it with elements, 
   *** and assign to each element an order based on its placement
   *** in the array
   ***/
  SetElement[] addWithOrder (Collection elements)
  {
    SetElement[] result = new SetElement [elements.size ()];
    int idx = 0;
    
    for (Iterator it = elements.iterator (); it.hasNext ();)
      {
	SetElement element = (SetElement)it.next ();
	result [idx] = element;
	element.setOrd (idx);
	idx++;
      }
    return result;
  }
  
  void constructLowerCovers (Collection aboveRelation)
  {
    for (Iterator it = aboveRelation.iterator (); it.hasNext ();)
      {
	Object[] above = (Object[])it.next ();
	SetElement a = getElementByKey (above [0]);
	SetElement b = getElementByKey (above [1]);
	// -- a > b so we put b into lower cover of a
	addToLowerCover (a, b);
      }    
  }

  Set getLowerCover (SetElement a)
  {
    Set lowerCover = (Set)lowerCovers.get (a);
    if (lowerCover == null) return Collections.EMPTY_SET;
    return lowerCover;
  }
  void addToLowerCover (SetElement a, SetElement b)
  {
    Set lowerCover = (Set)lowerCovers.get (a);
    if (lowerCover == null)
      {
	lowerCover = new HashSet ();
	lowerCovers.put (a, lowerCover);
      }
    lowerCover.add (b);
  }
  

  void constructUpperCovers (Collection aboveRelation)
  {
    // -- each element in the aboveRelation is of type Object[2]
    // -- such that aboveRelation [0] >= aboveRelation [1]
    for (Iterator it = aboveRelation.iterator (); it.hasNext ();)
      {
	Object[] above = (Object[])it.next ();
	SetElement a = getElementByKey (above [0]);
	SetElement b = getElementByKey (above [1]);
	// -- a >= b so we put a into upper cover of b
	
	addToUpperCover (b, a);
      }
  }
  
  Set getUpperCover (SetElement a)
  {
    Set upperCover = (Set)upperCovers.get (a);
    if (upperCover == null) return Collections.EMPTY_SET;
    return upperCover;
  }
  void addToUpperCover (SetElement a, SetElement b)
  {
    Set upperCover = (Set)upperCovers.get (a);
    if (upperCover == null)
      {
	upperCover = new HashSet ();
	upperCovers.put (a, upperCover);
      }
    upperCover.add (b);
  }
  

  /** build a map between the lattice elements as view by the user
   ** and our internal representation
   **/
  void mapElements (Collection _elements)
  {
    dataToElements = new HashMap ();
    int count = 0;
    for (Iterator it = _elements.iterator (); it.hasNext ();)
      {
	Object o = it.next ();
	dataToElements.put (o, new SetElement (this, count++, o));
      }
  }
  

  // -- sorts elements in reverse toplogical order 
  // -- (also known as linear extension)
  SetElement[] sortElements (Collection col)
  {
    // -- create axulary array filled with '0'
    int[] count = new int [col.size ()];
    Arrays.fill (count, 0);

    // -- fill count with numbers such that
    // -- given an element 'a'
    // -- the value of count [a.getOrd ()] is exactly the number 
    // -- of occurences of a in an upperCover of other elements
    for (Iterator it = col.iterator (); it.hasNext ();)
      {
	SetElement element = (SetElement)it.next ();
	Set upperCover = getUpperCover (element);
	for (Iterator it2 = upperCover.iterator (); it2.hasNext ();)
	  {
	    SetElement upCoverElement = (SetElement)it2.next ();
	    count [upCoverElement.getOrd ()]++;
	  }
      }
    
    // -- stack of elements that have to be processed
    Stack needProcessing = new Stack ();
    for (int i = 0; i < count.length; i++)
      if (count [i] == 0)
	needProcessing.push (getElement (i)); 

    // -- set up our result array
    SetElement[] result = new SetElement [count.length];
    int idx = count.length - 1;
    
    // -- while we have elements to process keep adding them to the result
    while (!needProcessing.empty ())
      {
	SetElement element = (SetElement)needProcessing.pop ();

	// -- add this element to the list
	result [idx] = element;
	element.setOrd (idx--);
	
	for (Iterator it = getUpperCover (element).iterator (); it.hasNext ();)
	  {
	    SetElement upCoverElement = (SetElement)it.next ();
	    count [upCoverElement.getOrd ()]--;
	    if (count [upCoverElement.getOrd ()] == 0)
	      needProcessing.push (upCoverElement);
	  }
      }
    return result;
  }
  
  /***
   *** Builds a join table 
   ***/
  void buildJoinTable ()
  {
    joinTable = new int [elements.length][elements.length];


    for (int i = 0; i < elements.length; i++)
      {	
	joinTable [i][i] = getElement (i).getOrd ();

	// -- compute join with all of the elements we have processed 
	// -- so far
	for (int j = 0; j < i; j++)
	  {
	    
	    int q = -1;
	    for (Iterator it = getUpperCover (getElement (i)).iterator ();
		 it.hasNext ();)
	      {
		/****
		 **** 
		 **** We need to compute q where q is the maximum of the set
		 **** { e.getOrd () | e = j \/ x and x \in upperCover[i] }
		 */
		SetElement upCoverElement = (SetElement)it.next ();
		if (joinTable [j][upCoverElement.getOrd ()] > q)
		  q = joinTable [j][upCoverElement.getOrd ()];
	      }
	    joinTable [i][j] = q;
	    joinTable [j][i] = q;
	  }
      }
  }
  
  /****
   **** Builds a meet table. This is dual to buildJoinTable and possibly
   **** we can encapsulate both constructions in a single method, but for 
   **** now they are separate
   ****/
  void buildMeetTable ()
  {
    meetTable = new int [elements.length][elements.length];


    // -- counts how many elements we've seen so far
    int idx = 0;
    for (Iterator it = new ArrayIterator (elements, false); 
	 it.hasNext ();)
      {
	SetElement x = (SetElement)it.next ();
	int ordX = x.getOrd ();

	meetTable [ordX][ordX] = ordX;

	// -- compute join with all of the elements we have processed 
	// -- so far
	//for (int j = 0; j < i; j++)
	for (Iterator it2 = new ArrayIterator (elements, false, idx); 
	     it2.hasNext ();)
	  {
	    SetElement y = (SetElement)it2.next ();
	    int q = size () + 1;
	    for (Iterator it3 = getLowerCover (x).iterator ();
		 it3.hasNext ();)
	      {
		/****
		 **** 
		 **** We need to compute q where q is the minimum of the set
		 **** { e.getOrd () | e = y /\ z and z \in upperCover[i] }
		 */
		SetElement z = (SetElement)it3.next ();
		if (meetTable [y.getOrd ()][z.getOrd ()] < q)
		  q = meetTable [y.getOrd ()][z.getOrd ()];
	      }
	    //System.out.println (x + " meet " + y + " = " + q);
	    meetTable [ordX][y.getOrd ()] = q;
	    meetTable [y.getOrd ()][ordX] = q;
	  }
	idx++;
      }

  }
  

  // -- given an opaque index of the ordered set element return the element data
  public Object getElementData (int idx)
  {
    return getElement (idx).getData ();
  }
  
  public SetElement getElement (int idx)
  {
    //System.out.println ("Getting element at: " + idx);
    return elements [idx];
  }
  
  public SetElement getElementByKey (Object key)
  {
    return (SetElement)dataToElements.get (key);
  }

  // -- since elements are sorted getting top and bottom is trivial
  public SetElement getTop ()
  {
    return elements [0];
  }
  
  public SetElement getBottom ()
  {
    return elements [elements.length - 1];
  }
  
  public int getTopIdx ()
  {
    return getTop ().getOrd ();
  }
  
  public int getBottomIdx ()
  {
    return getBottom ().getOrd ();
  }
  
  public int join (int a, int b)
  {
    return joinTable [a][b];
  }
  public int meet (int a, int b)
  {
    return meetTable [a][b];
  }


  SetElement[] getElements ()
  {
    return elements;
  }
  
  void constrcutProduct (OrderedSet set1, OrderedSet set2, 
			 Collection elements, 
			 Collection aboveRelation)
  {
    SetElement[] elements1 = set1.getElements ();
    SetElement[] elements2 = set2.getElements ();

    for (int i = 0; i < elements1.length; i++)
      {
	for (int j = 0; j < elements2.length; j++)
	  {
	    // -- new lattice value is a concatination of the 
	    // -- elements1[i] and elements2[j]
	    String latValue = elements1 [i].getData ().toString () + 
	      elements2 [j].getData ().toString ();
	    // -- add the value to our elements
	    elements.add (latValue);
	    // -- build the above realtions
	    buildAboveRelation (set1, set2, elements1 [i], elements2 [j],
				aboveRelation);
	    
	  }
      }
    
    
  }
  
  void buildAboveRelation (OrderedSet set1, OrderedSet set2, 
			   SetElement elm1, SetElement elm2, 
			   Collection aboveRelation)
  {
    Set lowerCover1 = set1.getLowerCover (elm1);
    Set lowerCover2 = set2.getLowerCover (elm2);
    
    for (Iterator it = lowerCover2.iterator (); it.hasNext ();)
      {
	SetElement low = (SetElement)it.next ();
	Object[] rel = new Object [2];
	rel [0] = elm1.getData ().toString () + elm2.getData ().toString ();
	rel [1] = elm1.getData ().toString () + low.getData ().toString ();
	aboveRelation.add (rel);
      }

    for (Iterator it = lowerCover1.iterator (); it.hasNext ();)
      {
	SetElement low = (SetElement)it.next ();
	Object[] rel = new Object [2];
	rel [0] = elm1.getData ().toString() + elm2.getData ().toString ();
	rel [1] = low.getData ().toString () + elm2.getData ().toString ();
	aboveRelation.add (rel);
      }

    for (Iterator it = lowerCover1.iterator (); it.hasNext ();)
      {
	SetElement low1 = (SetElement)it.next ();
	for (Iterator it2 = lowerCover2.iterator (); it2.hasNext ();)
	  {
	    SetElement low2 = (SetElement)it2.next ();
	    Object[] rel = new Object [2];
	    rel [0] = elm1.getData ().toString() + 
	      elm2.getData ().toString ();
	    rel [1] = low1.getData ().toString () +
	      low2.getData ().toString ();
	    aboveRelation.add (rel);
	  }
      }
    
  }
  

  // -- an element of an ordered set
  public static class SetElement
  {
    // -- order of this element in the topological ordering
    int ord;

    // -- the set to which this elements belongs 
    // -- this can be used to avoid mixing elements from different sets
    OrderedSet parentSet;

    // -- an element can have arbitrary data attached to it, such as its
    // -- symbolic name if necessary
    Object data;
    

    public SetElement (OrderedSet _parentSet, int _ord, Object _data)
    {
      data = _data;
      parentSet = _parentSet;
      ord = _ord;
    }
    
    
    public String toString ()
    {
      if (data != null) return data.toString ();
      return super.toString ();
    }
    
    public boolean equals (Object o)
    {
      if (! (o instanceof SetElement)) return false;
      return equals ((SetElement)o);
    }
    
    public boolean equals (SetElement element)
    {
      if (!element.getParentSet ().equals (getParentSet ())) return false;
      
      return element.getOrd () == getOrd ();
    }
    
    public int getOrd ()
    {
      return ord;
    }
    
    public void setOrd (int v)
    {
      ord = v;
    }
    
    public OrderedSet getParentSet ()
    {
      return parentSet;
    }
    
    public Object getData ()
    {
      return data;
    }
  }
  
}
