package edu.toronto.cs.ctl;

import java.util.*;
import java.io.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.tlq.*;
import edu.toronto.cs.util.caching.*;
import edu.toronto.cs.smv.*;

/**
 ** This class allows to convert CTLPlaceholderNode into something
 ** that can be used by the model checker and CTLMacroManager.
 **/
public class PlaceholderManager extends MacroExpander
{

  // -- list of term expansions for each placeholder
  // -- each element is of type CTLNode[]
  List placeHolderTerms;
  
  // -- name of all place holders that we know about
  // -- the expanded term for each placeholder are in 'terms', 
  // -- i.e. terms.get (i) correspond to placeHolders.get (i)
  List placeHolders;
  
  
  // -- the top free bit
  int topBit;
  
  
  UpSetLattice lat;

  EnumeratedTypes enumTypes;

  public PlaceholderManager ()
  {
  }
  

  /**
   ** Constructs a new placeholder manager.
   **/
  public PlaceholderManager (UpSetLattice lattice, EnumeratedTypes _enumTypes)
  {
    placeHolderTerms = new ArrayList ();
    placeHolders = new ArrayList ();
    
    lat = lattice;
    enumTypes = _enumTypes;
    topBit = 0;
    
  }

  public PlaceholderManager (UpSetLattice lattice)
  {
    this (lattice, EnumeratedTypes.EMPTY_TYPES);
  }
  

  public void reset ()
  {
    placeHolderTerms = new ArrayList ();
    placeHolders = new ArrayList ();
    topBit = 0;
  }
  
  public CTLNode expand (CTLNode ctl)
  {
    if (ctl instanceof CTLPlaceholderNode)
      return expandPlaceholder ((CTLPlaceholderNode)ctl);
    
    return ctl;
  }
  

  public UpSetLattice getLattice ()
  {
    return lat;
  }
  
  /**
   ** Expands placeholder and constructs a map from integers to the
   ** relevant propositional terms.
   **/
  public CTLNode expandPlaceholder (CTLPlaceholderNode ph)
  {
    System.out.println ("PlaceHolderManager: expandPlaceholder: " + ph);
    
    // -- remember when we first saw this place holder
    placeHolders.add (ph);

    String [] props = (String []) ph.getProps ();
    int n; // counts prop terms

    CTLNode templat;
    CTLNode tempprop;
    CTLNode result = null;



    // -- props with expanded enum vars
    String[] expandedProps;
    
    List expandedList = new ArrayList ();
    
    for (int i = 0; i < props.length; i++)
      {
	if (enumTypes.isEnumeratedVariable (props [i]))
	  for (int j = 0; j < enumTypes.getRangeBits (props [i]); j++)
	    expandedList.add (props [i] + j);
	else
	  expandedList.add (props [i]);
      }

    expandedProps = 
      (String[])expandedList.toArray (new String [expandedList.size ()]);

    // -- 2 ^ props.length
    CTLNode[] terms = new CTLNode [1 << expandedProps.length];


    for (n = 0; n < terms.length; n++)
      {

	// -- get the lattice value for the next free bit
	templat = 
	  CTLFactory.createCTLLatticeNode (lat.getUpMinTerm (topBit + n),
					    lat);

	tempprop = makePropFor (n, expandedProps);

	// -- if we have enumerated types we should 
	// -- store the formula corresponding to that term using
	// -- the enumerations
	if (enumTypes != null)
	  terms [n] = makeEnumTerm (n, props);
	else
	  terms [n] = tempprop;

	if (result == null)
	  result = tempprop.and (templat);
	else 
	  result = result.or (tempprop.and (templat));
      }
    
    // -- update topBit
    topBit += terms.length;
    
    placeHolderTerms.add (terms);
    return result;
  }


  // -- given a lattice value interpret it using propositional formulas
  public List interpretSoln (int latVal)
  {
    List solution;
    if (latVal == lat.idtimesIdx () || latVal == lat.idplusIdx ())
      {
	solution = new LinkedList ();
	solution.add (CTLFactory.createCTLLatticeNode (latVal, lat));
      }
    else
      solution = interpretSoln ((int[]) lat.getElementKeyByIdx (latVal));

    return solution;
  }
  
  /**
   ** Given an array of integers returns their interpretation as a
   ** List of solutions.
   **/
  public List interpretSoln (int [] soln)
  {
    List solution = new LinkedList ();
    
    for (int i = 0; i < soln.length; i++)
      solution.add (intToSoln (soln [i]));

    return solution;
  }



  CTLNode intToSoln (int sol)
  {
    int length = placeHolderTerms.size ();
    CTLNode result = null;

    for (int i = 0; i < length; i++)
      {
	
	CTLNode[] terms = (CTLNode[])placeHolderTerms.get (i);
	CTLNode tempResult = (CTLNode)placeHolders.get (i);
	tempResult = tempResult.eq (intToSoln (sol, terms));
	
	if (result == null)
	  result = tempResult;
	else
	  result = result.and (tempResult);

	// -- removed bits already processed
	sol = sol >>> terms.length;
      }
    return result;
  }
  
  CTLNode intToSoln (int sol, CTLNode[] terms)
  {
    CTLNode solution = null;
    int bits = sol;
    
    // -- the bit we are curently considering
    for (int nbit = 0; bits != 0 && nbit < terms.length; nbit++)
      {
	if ((bits & 1) == 1)
	  {
	    if (solution == null) 
	      solution = terms [nbit];
	    else 
	      solution = solution.or (terms [nbit]);
	  }

	// -- remove the bit we've processed
	bits = bits >>> 1;
      }

    // -- it is possible that solution is still null, if 
    // -- for this place holder anything (including F) is 
    // -- a solution, so we hack it up a bit to see the desired
    // -- output
    if (solution == null)
      solution =
	CTLFactory.createCTLLatticeNode (lat.idtimesIdx (), lat);
    return solution;
  }
  

  /**
   ** Creates the prop formula (term) corresponding to the number n.
   **/
  private CTLNode makePropFor (int n, String [] props)
  {
    CTLNode result;
    int bits = n;

    if ((bits & 1) == 1) // 0th bit is on -> p[0]
      result = CTLFactory.createCTLAtomNode (props [0]);
    else // -> !p[0]
      result = CTLFactory.createCTLAtomNode (props [0]).neg ();

    // -- shift bits since we've handled the 0 bit
    bits = bits >>> 1;

    // cycle through the rest of the atomic props 
    // and build a large conjunction
    for (int i = 1; i < props.length; i++)
      {
	if ((bits & 1) == 1) // 0th bit is on -> result /\ p[i]
	  result = result.and
	    (CTLFactory.createCTLAtomNode (props [i]));
	else // -> result /\ !p[i]
	  result = result.and 
	    (CTLFactory.createCTLAtomNode (props [i]).neg ());
	bits = bits >>> 1;
      }
    return result;
  }

  // -- n is the current encoding of a cube
  // -- props is the set of all variables we are delaing with
  // -- in the order of the encoding. i.e. props [0] is a variable correspoding
  // -- to bits 0..k where k is number of bits needed to encode props [0]
  // -- props [1] starts from bits k+1, etc
  CTLNode makeEnumTerm (int n, String[] props)
  {
    CTLNode result = null;

    int bitsSeen = 0;
    

    for (int i = 0; i < props.length; i++)
      {
	int width = enumTypes.getRangeBits (props [i]);
	
	if (result == null)
	  result = makeEnumTerm (getBits (n, bitsSeen, width), props [i]);
	else
	  result = result.
	    and (makeEnumTerm (getBits (n, bitsSeen, width), props [i]));
	bitsSeen += width;
      }
    

    return result;
  }

  CTLNode makeEnumTerm (int n, String varName)
  {

    CTLNode varNode = CTLFactory.createCTLAtomNode (varName);
    // -- take care of non enumerated variables
    if (!enumTypes.isEnumeratedVariable (varName))
      {
	if (n == 0) 
	  varNode = varNode.neg ();
	return varNode;
      }
    

    // -- and now for the enumerated variables
    String enumVal = enumTypes.valueEncoded (varName, n);
    return varNode.eq (CTLFactory.createCTLAtomNode (enumVal));
  }
  

  public static final int MAX_BITS = 32;
  
  // -- returns 'width' many bits of n starting from offset
  int getBits (int n, int offset, int width)
  {
    int result = n;
    
    // -- first shift it offset bits to the right to get rid of 
    // -- stuff we don't want
    result = result >>> offset;
    
    // -- now shift to the left and back tot he right
    // -- to get rid of top bits
    result = result << MAX_BITS - width;
    result = result >>> MAX_BITS - width;
    return result;
  }
  
  
}


