package edu.toronto.cs.experiments;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.util.*;


public class MvPriorityExperiment
{

  // -- number of cubes to alter
  int noiseLevel;
  FiniteLattice lattice;

  int top;
  int bot;

  ChaosIterator varDomain;

  ExperimentOutput output;
  
  // -- number of ftoSize we are merging
  int ftoSize;

  String ddFileName;
  
  public MvPriorityExperiment (FiniteLattice _lattice, 
			      ExperimentOutput _output, 
			      int _ftoSize,
			      String _ddFileName,
			      int _noiseLevel)
  {
    
    lattice = _lattice;
    output = _output;
    ftoSize = _ftoSize;
    ddFileName = _ddFileName;
    noiseLevel = _noiseLevel;

    top = lattice.idtimesIdx ();
    bot = lattice.idplusIdx ();
    
    varDomain = new UniformChaosIterator (new int[] {top, bot});
    
    
  }
  

  public void runExperiment ()
  {

    // -- create our manager XXX ignore arrity
    MDDManager manager = new MDDMgr (0, lattice);
    

    // -- get our join irreducibles
    int[] joinIrr = lattice.getJoinIrreducibles ();
    
    MDDNode[] mbtddNodes = new MDDNode [joinIrr.length];
    
    // -- first view point comes from a file
    MDDNode bddNode = DDUnDumper.readBDDSafe (manager, ddFileName);

    // -- priorities the original diagram
    MDDNode mddNode = introduceMvNoise (manager, bddNode);

//     Assert.assert (mddNode != bddNode);

    // -- mdd sample is ready
    output.addSample ("mdd", manager.dagSize (mddNode));
    

    // -- create mbtddNodes now
    for (int i = 0; i < mbtddNodes.length; i++)
      {
	//System.out.println ("Perfoming a cut for: " + joinIrr [i]);
	// -- Let f be an mdd, and let h be its mbtdd encoding, then
	// -- h_i = f \above i
	mbtddNodes [i] = manager.Apply (mddNode, 
					manager.getConstant (joinIrr [i]),
					FiniteLattice.ABOVE);
      }
    

//     Assert.assert (mbtddNodes [0] != mddNode);
    
//     Assert.assert (mbtddNodes [0] != mbtddNodes [1]);
      

    // -- we are ready for the 'mbtdd' case
    output.addSample ("mbtdd", manager.sharedSize (mbtddNodes));
    output.flush ();
  }
  

			      
  public int getRandomNegativeValue ()
  {
    // -- pick a lattice value between [M, \bot)
    int middle = (lattice.size ()) >> 1;

    int lo;
    int hi;
    
    if (middle < bot)
      {
	lo = middle;
	hi = bot;
      }
    else
      {
	// -- in range includes lower bound, but not the high bound
	// -- i.e. [lo, hi)
	lo = bot + 1;
	hi = middle + 1; 
      }

    
    int val = Chaos.inRange (lo, hi);
    // System.out.println ("Decided on a val: " + val);
    return val;
  }
  
  // -- introduce mv-noise into a decision diagram
  public  MDDNode introduceMvNoise (MDDManager manager, MDDNode node)
  {
    // -- given a decision diagram, and an FTO with the middle value M
    // -- we want to map some \top cubes to the interval (\top, M]
    // -- and some \bot cubes to the interval [M, \top)
    MDDNode positiveCubes = manager.getZero ();
    MDDNode negativeCubes = manager.getZero ();
    
    // -- find cubes we will change
    for (int i = 0; i < noiseLevel; i++)
      {
	// -- get a random poin
	MDDPoint point = manager.getRandomPoint (node, varDomain);
	if (point.getValue () == top)
	  {

	    // -- got a positive cube that we want to augment
	    
	    // -- get a random negative value i.e. in interval [M, \bot)
	    // -- since we will negate everything later on anyways
	    int val = getRandomNegativeValue ();

	    MDDNode pointNode = point.getNode ();
	    
	    // -- set the value
	    pointNode = manager.setEqualTo (pointNode, val);
	    
	    
	    // -- add to the rest of our cubes
	    positiveCubes = manager.Apply (positiveCubes, pointNode, 
					   FiniteLattice.JOIN);
	  }
	else
	  {
	    int val = getRandomNegativeValue ();
	    MDDNode pointNode = point.getNode ();
	    pointNode = manager.setEqualTo (pointNode, val);
	    
	    negativeCubes = manager.Apply (negativeCubes, pointNode,
					   FiniteLattice.JOIN);
	  }
	
      }
    

    MDDNode result = node;

    System.out.println ("Merging all negative cubes...");
    // -- change all of the negative cubes    
    result = manager.Apply (result, negativeCubes, 
			    FiniteLattice.JOIN);
    
    System.out.println ("Merging all positive cubes....");
    // -- change all positive cubes
    result = manager.Apply 
      (result, manager.Apply (positiveCubes, FiniteLattice.NEG),
       FiniteLattice.MEET);
    return result;
  }
  

  public static void main (String[] args) throws Exception
  {
    // -- intialize seed so that we could re-run this
    Chaos.seed (0);
    
    String ddFileName = args [0];
    String outputDir = args [1];
    String experimentName = args [2];


    int ftoSize = Integer.parseInt (args [3]);
    int noiseLevel = Integer.parseInt (args [4]);
    
    
    ExperimentOutput output = 
      new ExperimentOutput (outputDir, experimentName);
    FiniteLattice lattice;

    
    int latticeSize = ftoSize;
    
    lattice = Lattices.getLatticeByName (latticeSize + "FTO");
    System.out.println ("Getting a lattice: " + latticeSize + "FTO");

    for (int i = 0; i < 100; i++)
      {
	System.out.println ("Getting sample: " + i);
	MvPriorityExperiment experiment = 
	  new MvPriorityExperiment (lattice, 
				    output, 
				    ftoSize, 
				    ddFileName,
				    noiseLevel);
	System.out.println ("Running experiment....");
	experiment.runExperiment ();
	System.out.println ("Done");
      }
    
  }
  
}
