package edu.toronto.cs.mdd;

import java.io.*;
import java.util.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;


import edu.toronto.cs.experiments.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.modelchecker.*;


public class DDUnDumper
{
  MDDManager manager;
  MarshaledInputStream in;

  // -- nodes that we've already seen
  List seen;

  String logicName;
  int arrity;
  
  MDDNode currentNode;
  boolean complement = false;
  
  public DDUnDumper (InputStream _in, MDDManager _manager)
  {
    in = new MarshaledInputStream (_in);
    manager = _manager;
  }
  public DDUnDumper (MarshaledInputStream _in, MDDManager _manager)
  {
    in = _in;
    manager = _manager;
  }


  public MDDNode readDD () throws IOException
  {
    seen = new ArrayList ();
    // -- occupy 0 position to force indexes to start from 1
    seen.add (new Object ());
    
    readHeaders ();

    // -- read until end-of-file is reached
    try {
      while (true)
	currentNode = readDDNode ();
    }
    catch (EOFException e) 
      {
	; // -- nothing to do here
      }
    
    return complement ? manager.Apply (currentNode, FiniteLattice.NEG) : 
      currentNode;
  }
  
  void readHeaders () throws IOException
  {
    if (in.readChar () != 'D' ||
	in.readChar () != 'D')
      throw new IOException ("Malformed format, does not start with DD");
    
    if (in.readByte () != DDDumper.LATTICE_NAME)
      throw new IOException ("Malformed format, no lattice name");
    
    logicName = in.readString ();
    
    int dataType = in.readByte ();

    
    if (dataType == DDDumper.DD_DATA_COMPELEMENT)
      complement = true;
    else if (dataType == DDDumper.DD_DATA)
      complement = false;
    else
      throw new 
	IOException ("Malformed format, expecting DD_DATA(_COMPLEMENT)");
    
    arrity = in.readShort ();
   
  }
  
  MDDNode readDDNode () throws IOException
  {
    // -- read the node index
    int varIndex = in.readInt ();

    if (varIndex == -1) throw new EOFException ("Done");
    
			  
    // -- read its children
    int[] children = in.readIntArrayFixed (arrity);


    children = ArrayUtil.reverse (children);

    MDDNode result;

    // -- node is constant
    if (children [0] == DDDumper.NO_CHILD) 
      result = manager.getConstant (varIndex);
    
    
    else // -- node is not constant
      {
	// -- resolve all children
	MDDNode[] ddChildren = new MDDNode [arrity];
	for (int i = 0; i < children.length; i++)
	  {
	    ddChildren [i] = (MDDNode)seen.get (children [i] >> 1);
	    // -- negate odd children
	    if ((children [i] & 1) != 0) {
	      //	      System.out.println("Processing complement edge to "+(children[i]>>1));
	      
	      ddChildren [i] = 
		manager.Apply (ddChildren [i], FiniteLattice.NEG);
	    }
	    
	  }
	
	// -- create result
	//System.out.println("making node for varIndex="+varIndex);
	
	result = manager.makeUnique (varIndex, ddChildren);
      }  
    seen.add (result);
    return result;
  }

  public static MDDNode readBDDSafe (MDDManager manager, String fileName)
  {
    try
      {
	return readBDD (manager, fileName);
      }
    catch (Exception ex) 
      {
	Assert.assert (ex);
      }
    return null;
  }
  
  // -- reads a BDD from a file and imports it into a given 
  // -- MDDManager
  public static MDDNode readBDD (MDDManager manager, String fileName)
    throws IOException
  {
    FiniteLattice boolLattice = Lattices.getLatticeByName ("2Bool");

    // XXX We assume that arrity is not used anywhere and thus it
    // XXX is set to 0
    MDDManager boolManager = new MDDMgr (0, boolLattice);
    
    DDUnDumper ddReader = new DDUnDumper (new FileInputStream (fileName),
					  boolManager);
    
    // -- read in the decision diagram
    System.out.println ("Reading BDD");
    MDDNode bdd = ddReader.readDD ();
    
    
    FiniteLattice lattice = (FiniteLattice)manager.getAlgebra ();

    // -- create a logic map
    int[] logicMap = new int [2];
    logicMap [boolLattice.idtimesIdx ()] = lattice.idtimesIdx ();
    logicMap [boolLattice.idplusIdx ()] = lattice.idplusIdx ();
    

    System.out.println ("Transfering BDD");
    // -- transfer the diagram to the new manager
    return boolManager.transfer (manager, bdd, logicMap);
    
  }
  

  public static MvSet readBDDSafe (MvSetFactory mvSetFactory, String fileName)
  {
    try
      {
	return readBDD (mvSetFactory, fileName);
      }
    catch (Exception ex) 
      {
	Assert.assert (ex);
      }
    return null;
  }


  public static MvSet readBDD (MvSetFactory mvSetFactory, String fileName)
    throws IOException
  {
    FiniteLattice boolLattice = Lattices.getLatticeByName ("2Bool");

    // XXX We assume that arrity is not used anywhere and thus it
    // XXX is set to 0
    MDDManager boolManager = new MDDMgr (0, boolLattice);
    
    DDUnDumper ddReader = new DDUnDumper (new FileInputStream (fileName),
					  boolManager);
    
    // -- read in the decision diagram
    System.out.print ("Reading BDD ... ");
    MDDNode bdd = ddReader.readDD ();

    System.out.println (boolManager.dagSize (bdd));

    
    FiniteLattice lattice = mvSetFactory.getLattice ();

    // -- create a logic map
    int[] logicMap = new int [2];
    logicMap [boolLattice.idtimesIdx ()] = lattice.idtimesIdx ();
    logicMap [boolLattice.idplusIdx ()] = lattice.idplusIdx ();
    

    System.out.println ("Transfering BDD");
    // -- transfer the diagram to the new manager
    return boolManager.transfer (mvSetFactory, bdd, logicMap);
  }
  


  public static void main (String[] args)
  {
    String filename = args [0];
    MvSetFactory factory = 
      MDDMvSet.newMvSetFactory (Lattices.getLatticeByName ("2Bool"), 50);
    MvSet mvSet = readBDDSafe (factory, filename);
    PropOrderIf ordering = new AlternatePropOrder (new String[]{"b","a"});

    mvSet.writeToDaVinci ("/tmp/blat.daVinci", ordering);

    System.out.println ("The size is: " + mvSet.size ());
    System.out.println ("Reordering...");
    mvSet.reorder ();
    System.out.println ("Size after reoredering: " + mvSet.size ());
  }
  
  public static void _main (String[] args)
  {
    String fileName = args [0];
    
    MDDManager manager = new MDDMgr (0, Lattices.getLatticeByName ("4Bool"));
    
    MDDNode node = readBDDSafe (manager, fileName);
    
  }
  
}
