package edu.toronto.cs.modelchecker;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.userprompt.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.prettygraph.*;
import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.xkripke.XKripke.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.smv.*;

import edu.toronto.cs.tlq.UpSetLattice;

import java.util.*;
import java.io.*;
import java.lang.reflect.Method;

/**
 ** This is the manager for the model-checker. It sets up the
 ** modelchecker according to the specified options.
 **/
public class MCManager
{

  // initialize the logger
  static Logger log = new Logger (MCManager.class.getName ());

  // the xkripke structure(s)
  XKripke[] xkripke;

  // the propositional ordering
  PropOrderIf propOrder;

  // the associated model-checker
  MvSetModelChecker mc;
  
  // various options
  Map options;

  // verbosity level: 0 - lowest, 2 - highest (deafult)
  int verbose;

  // initial state
  MvState initState;

  // model reader to be used
  AbstractModelReader modelReader = null;

  // the macro expander for handling state names
  //  StringMacroManager macro;
  CTLMacroManager macroManager;

  PlaceholderManager2 placeHolderManager = null;

  // -- enumerated types that we know about
  EnumeratedTypes enumTypes = EnumeratedTypes.EMPTY_TYPES;

  // list of fairness conditions
  List fairness = new LinkedList ();

  /**
   ** Initialises the modelchecker front end with the specified options.
   ** 
   ** @param opts -- a Map of various options
   **/
  public MCManager (Map opts)
  {
    // parse the command line options
    options = opts;

    log.info ("Options passed to the modelchecker: " + options);

    // get the verbosity level: 0 - lowest, 2 - highest (default)
    verbose = ((Integer) options.get ("verbose")).intValue ();
    if (verbose < 0)
      verbose = 0;
    else if (verbose > 2)
      verbose = 2;

    MvSet transRelation = null;
    
    // XXX This was not thought through, but it will WORK! hopefully
    
    if (((Boolean)options.get ("smv")).booleanValue ())
      {
	try {
	  modelReader = new SMVModelReader (options);
	} catch (Exception ex) {
	  ex.printStackTrace ();
	  throw new IllegalArgumentException ("Some error or other");
	}
      } 
    // B.D. : kludge to use DDUnDumper to load in a transition system.
    else if (((Boolean)options.get("bdd")).booleanValue())
      {
      
	log.error ("You choose poorly.");
      
	// get the BDD file name
	// get the variable ordering file name
	// load the BDD file, make a transition relation
	// load the variable-ordering file, make a propositional ordering
	return;
      }
    else
      {
	modelReader = new XMLModelReader (options);
      }
    xkripke = modelReader.getXKripke ();
    propOrder = modelReader.getPropOrder ();

    // now we can create an MvSetFactory
    // get the lattice -- assumes that all the lattices used are the same
    FiniteLattice lattice = modelReader.getLattice ();
      // xkripke [0].getLattice ();
    // create the mvset factory
    MvSetFactory mvSetFactory = createMvSetFactory (lattice);

    // make the transition relation for the model
    modelReader.makeTransRelation (mvSetFactory);
    // now we can initialize transRelation and initState
    transRelation = modelReader.getTransRelation ();
    initState = modelReader.getInitState ();

    // create the model-checker
    mc = new MvSetModelChecker (transRelation,
				propOrder, getCTLSimplifier ());

    enumTypes = modelReader.getEnumTypes ();

    setupMacroManagers ();

    System.out.println ("Some statistics: ");
    System.out.println ("Size of trans: " + transRelation.size ());
    //     System.out.println ("Saving transition relation to a file");
    //     transRelation.writeToDaVinci ("/tmp/transrel.daVinci", propOrder);
    //     transRelation = mc.computeRechable (initState);
    //     System.out.println ("Size of reachable trans: " + transRelation.size ());
    //     if (transRelation.size () < mc.getMvTransRelation ().size ())
    //       mc.setTransitionRelation (transRelation);


    
    //     transRelation.reorder ();
    //     System.out.println ("Size of reachable trans after reorder: " +
    //  			transRelation.size ());

  }

  /**
   ** Set up the macro managers.
   **/
  private void setupMacroManagers ()
  {
    // -- create a macro manager
    macroManager = new CTLMacroManager (mc.getLattice ());
    
    // -- if possible create a place-holder manager
    if (mc.getLattice () instanceof UpSetLattice)
      {
	placeHolderManager = 
	  new PlaceholderManager2 ((UpSetLattice) mc.getLattice (),
				   enumTypes);
	macroManager.addExpander (placeHolderManager);
      }

    // -- we have enumerated types, so create macroExpander for them
    if (enumTypes != null && enumTypes != EnumeratedTypes.EMPTY_TYPES)
      macroManager.addExpander (new SMVEnumTypeExpander (enumTypes));
  }

  /**
   ** Pass the option map to the MCManager.
   **/
  public void passOptions (Map opts)
  {
    options = opts;
  }
  
  private CTLSimplifier getCTLSimplifier ()
  {
    if (options.get ("expansion").equals ("e"))
      return new CTLEXEUEGSimplifier ();
    else if (options.get ("expansion").equals ("a"))
      return new CTLDirectSimplifier ();
    else throw new RuntimeException ("Incorrect choice of CTL simplifier :" +
				     "--expansion " +
				     options.get ("expansion"));
  }

  /**
   ** Creates an MvSetFactory according to the options.
   **/
  private MvSetFactory createMvSetFactory (FiniteLattice lattice)
  {
    // get the MvSet type
    String mvsettype = (String) options.get ("mvset");

    // select MvSetFactory
    if (mvsettype.equals ("mdd") || mvsettype.equals ("MDDMvSet"))
      // -- MDD based
      {
	log.debug ("Using MDDMvSets");
	return MDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("join") || mvsettype.equals ("JMvSet"))
      // -- Join Irreducible based
      {
	log.debug ("Using JMvSets");
	return JMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("jbdd") || mvsettype.equals ("JBDDMvSet"))
      // -- JBDD implementation
      {
	log.debug ("Using JBDDMvSets");
	return JBDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }   
    else if (mvsettype.equals ("cubdd") || mvsettype.equals ("CUBDDMvSet"))
      {
	log.debug ("Using CUBDDMvSet");
	return CUBDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("cuadd") || mvsettype.equals ("CUADDMvSet"))
      {
	log.debug ("Using CUADDMvSet");
	return CUADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("jadd") || mvsettype.equals ("JADDMvSet"))
      {
	log.debug ("Using JADDMvSet");
	return JADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("cubiadd") || mvsettype.equals("CUBiADDMvSet"))
      {
	log.debug ("Using CUBiADDMvSet");
	
	return CUBiADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else throw new IllegalArgumentException 
	   ("MvSet type '" + mvsettype + "' not supported."); 
  }

  /**
   ** Gets the initial state for the current model. WARNING: only
   ** works with an XML model.
   **
   ** @return -- the initial state for the model or null if no unique
   ** initial state was defined.
   **/
  public MvState constructInitialState ()
  {
    modelReader.setOptions (options);
    ((XMLModelReader)modelReader).constructInitialState ();
    return modelReader.getInitState ();
  }

  /**
   ** Restricts the result to a unique initial state if one exists,
   ** otherwise does nothing.
   **/
  public MvSet restrictToInitialState (MvSet result)
    {
      // -- restrict the result to a unique initial state if one exists
      if (initState == null)
	return result; // do nothing
      else // restrict the result to initState
	{
// 	  System.out.println ("Restricting with respect to: ");
// 	  System.out.println (initState.toString (propOrder));
	  return restrictMvSet (initState, result);
	}
    }

  /**
   ** Given an MvState restricts the MvSet to it in order to obtain a
   ** single answer.
   **/
  private MvSet restrictMvSet (MvState state, MvSet mvSet)
    {
      return mvSet.restrict (state);
      //return result;
    }

  /**
   ** Given an MvSet result produce an approprite String representation.
   **/
  public String resultToString (MvSet result)
  {
    String out = "";

    // -- restrict the result to an initial state, this should result
    // -- in a constant MDD, i.e. a simple T, F, etc. answer
    if (verbose > 0)
      out = out + "Restricting to initial state\n";
    result = restrictToInitialState (result);

    // get the lattice value of the result
    int resultValue = result.evaluate
      (new MvState (new int [0], mc.getLattice ()));

    if (verbose > 1)
      out = out + "The result is: " + resultValue + "\n";

    // -- if we have a place holder manager, use it to interpret the
    // -- solution
    if (placeHolderManager == null)
      {
        out = out + "Final result: "
	  + mc.getLattice ().getElementKeyByIdx (resultValue)
	  + "\n";
      }
    else
      {
        out = out + "Final result:\n";
        List solution = placeHolderManager.interpretSoln (resultValue);
        int count = 0;
        for (Iterator it = solution.iterator (); it.hasNext ();)
          {
            out = out + "Solution " + (count++) + ": " + it.next () + "\n";
          }
      }

    return out;
  }

  /**
   ** Gets the model-checker.
   **/
  public MvSetModelChecker getMC ()
  {
    return mc;
  }

  /**
   ** Get the value of enumTypes.
   ** @return value of enumTypes.
   **/
  public EnumeratedTypes getEnumTypes()
  {
    return enumTypes;
  }

  /**
   ** Get the value of propOrder.
   ** @return value of propOrder.
   **/
  public PropOrderIf getPropOrder()
  {
    return propOrder;
  }

  /**
   ** Get the XKripke structure(s).
   ** @return XKripke structure(s).
   **/
  public XKripke [] getXKripke()
  {
    return xkripke;
  }
  
  /**
   ** Get the initial state.
   ** @return initial state.
   **/
  public MvState getInitState ()
  {
    return initState;
  }
  
  /**
   ** Gets the level of verbosity.
   ** @return the level of verbosity.
   **/
  public int getVerbosity ()
  {
    return verbose;
  }

  /**
   ** Get the CTL macro manager.
   ** @return the CTL macro manager.
   **/
  public CTLMacroManager getCTLMacroManager ()
  {
    return macroManager;
  }

  /**
   ** Set fairness.
   **/
  public void setFairness (List fair)
  {
    fairness = fair;
  }
  /**
   ** Get fairness.
   ** @return fairness.
   **/
  public List getFairness ()
  {
    return fairness;
  }
  /**
   ** Set the CTL macro manager.
   **/
  public void setCTLMacroManager (CTLMacroManager mm)
  {
    macroManager = mm;
  }

  /**
   ** Get the placeholder manager.
   ** @return the placeholder manager.
   **/
  public PlaceholderManager2 getPlaceHolderManager ()
  {
    return placeHolderManager;
  }
  /**
   ** Set the placeholder manager.
   **/
  public void setPlaceHolderManager (PlaceholderManager2 pm)
  {
    placeHolderManager = pm;
  }

}





















