package edu.toronto.cs.modelchecker;

import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.xkripke.XKripke.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.prettygraph.*;


import edu.toronto.cs.ctl.*;
import edu.toronto.cs.util.caching.*;

import edu.toronto.cs.mvset.*;

import java.io.*;
import java.util.*;

/****
 **** this may become an MDD model checker
 ****/
public class MvSetModelChecker
{
  // -- initializing the logger
  static Logger log = new Logger(MvSetModelChecker.class.getName());
  // static Logger special = new Logger("Special");
  // -- XKripke model to model check
  XKripke kripke;
  // -- CTL property we are verifying

  FiniteLattice lattice;
  // -- MvSet encoding of the kripke structure
  MvSet mvTransRelation;
   
  PropOrderIf propOrdering;
  
  // -- this is going to be used to cache stuff for counter examples
  Cache cache;

  // -- factory to create mvSets
  MvSetFactory mvSetFactory;

  int top, bottom;


  /****
   **** A trivial main for the model checker
   ****/
  public static void main (String[] args) throws Exception     
  {

    // -- parse the command line arguments
    Map opts = McClapi.parseOpts (args);

    // -- create the model checker
    // -- note that there is a List of input files, for now we only
    // consider the first element in that list
    MvSetModelChecker modelChecker = 
      new MvSetModelChecker (XKripkeFactory.parse 
			     ((String) ((List) opts.get ("input")).get (0)),
			     // new CTLEXEUEGSimplifier ());
			     new CTLDirectSimplifier ());

    CTLNode ctl = null;

    // -- prepare to read from stdin
    BufferedReader in = 
      new BufferedReader (new InputStreamReader (System.in));

    boolean stop = false; // -- stops the loop of inputs....

    do
      {
	// -- read a ctl formula
	System.out.println ("Please enter a CTL property");
	String ctlStr = in.readLine ();

	if (!ctlStr.equals (""))
	  {
	    System.out.println ("Parsing: " + ctlStr);    

	    // -- tell the converter what lattice it should use
	    BinaryTreeToCTLConverter.AssignLattice
	      (modelChecker.getLattice ());
	    // -- convert the ctlStr into a CTL tree
	    ctl = BinaryTreeToCTLConverter.convertToCTL (ctlStr);

	    // -- run our model checker on the property
	    MvSet mvResult = modelChecker.checkCTL (ctl);
	    MDDNode result = null;
	    if (mvResult instanceof MDDMvSet)
	      result = ((MDDMvSet)mvResult).getMddNode ();

	    // -- print some text representation of resulting mdd

	    System.out.println ("This is the formula that you are checking: " 
				+ ctl);
	    if (result != null)
	      {
		// -- write a graphical representation of resulting mdd
		System.out.println ("AM SO writing mdd to mdd.daVinci");
		//MDDNodeToDaVinci.mddToDaVinci 
		//  (result, modelChecker.propOrdering, "mdd.daVinci");

		// Below should work, but does (or at least 1) creates output and 2) doesn't crash).
  		DGraph tmp = PrettyGraphMDDVisitor.getPrettyRep 
  		  (result, modelChecker.propOrdering);
  		DVGenNodeVisitor dvg = new DVGenNodeVisitor 
  		  (new DVGenEdgeVisitor ());
  		for (Iterator it = DGDepthTraverse.getIterator (tmp);
 		     it.hasNext ();)
  		  {
  		    dvg.visit ((DGNode)it.next (), null);
  		  }
  		StringBuffer sb = dvg.getSB ();
  		FileWriter dvmdd = new FileWriter ("mdd.daVinci");
  		dvmdd.write (sb.toString ());
  		dvmdd.close ();
		
	      }
	    
    
	    // -- restrict the result to an initial state, this should result 
	    // -- in a constant MDD, i.e. a simple T, F, etc. answer
	    System.out.println ("Restricting to initial state");
	    //result = modelChecker.restrictToInitialState (result);
	    mvResult = modelChecker.restrictToInitialState (mvResult);

	    //result = ((MDDMvSet)mvResult).getMddNode ();
	    
	    // -- print the resulting restriction
	    //System.out.println (result);

	    // -- if we got a constant result (as we always should), find what 
	    // -- is the appropriate visual representation for this logic value
	    // -- is and show it to the user
//  	    if (result.isConstant ())
//  	      {
//  		// -- figure out what is the 'letter' value
//  		System.out.println ("Final Result: " + 
//  				    modelChecker.lattice.getElementKeyByIdx 
//  				    (result.getValue ()));
//  	      }
//  	    else
//  	      System.out.println ("WARNING: The result is not constant!");
	    
	    int resultValue = mvResult.evaluate 
	      (new MvState (new int [0], modelChecker.getLattice ()));
	    System.out.println ("Final result: " + 
				modelChecker.getLattice ().
				getElementKeyByIdx (resultValue));
    
	    // -- print property order, to help interpreting the result
	    modelChecker.printPropOrdering ();
	    PrintWriter out = new PrintWriter (System.out);
	    // -- print a map that maps user's idea about logic values and 
	    // -- internal representation used in debug print outs
	    modelChecker.lattice.dumpValueMap (out);

	    log.debug("The cache contains the following :");
	    Iterator i =  modelChecker.cache.keySet ().iterator ();
	    while (i.hasNext ())
	      log.debug(" -- " + i.next());
	    out.flush ();
	  }
	else stop = true;
      }
    while (!stop);
    // -- that's it, we are done
  }

  

  public MvSet restrictToInitialState (MvSet result)
    {
      // -- restrict the result to a unique initial state
      XKripkeState[] states = kripke.getInitialStates ();
      XKripkeState initState;
    
      if (states.length < 1)
	{
	  System.out.println ("NO initial states!");
	  return result;
	}
      else 
	{
	  initState = states [0];
	  if (states.length > 1) 
	    System.out.println ("WARNING: No unique initial state");
	}

      System.out.println ("Restricting with respect to: " + initState);
      return restrictMvSet (initState, result);
    
    }
  
  public MvSet restrictMvSet (XKripkeState state, MvSet mvSet)
    {
      MvSet result = mvSet;
      MvState mvState = new MvState (state.getProps ().size (), lattice);

      for (Iterator it = state.getProps ().values ().iterator ();
	   it.hasNext ();)
	{
	  XKripkeProp prop = (XKripkeProp)it.next ();
	  int varIndex = propOrdering.getPropOrder (prop.getName ());
	  int varValue = lattice.getElementIdxByKey (prop.getValue ());
	  
	  result = result.restrict (varIndex, varValue);
	  log.debug ("Restricting "+varIndex+" to "+varValue);
	}
      return result;
    }
  
  public void printPropOrdering ()
    {
      int size = propOrdering.size ();
      for (int i = 0; i < size; i++)
	{
	  System.out.print ("#" + i + " = " + propOrdering.getPropName (i));
	  if (propOrdering.isPostImage (i)) 
	    System.out.print ("'");
	  System.out.println ();
	}
    
    }

  public PropOrderIf getPropOrdering ()
  {
    return propOrdering;
  }


  /***
   *** MvSetModelChecker 
   *** XKripke -- a kripke structure to be model checked
   ***/
  public MvSetModelChecker (XKripke _kripke)
  {
    this (_kripke, new CTLDirectSimplifier ());
  }
  public MvSetModelChecker (XKripke _kripke, CTLSimplifier simplifier)
  {
    kripke = _kripke;
    //kripke.labeliseStates ();

    lattice = kripke.getLattice ();

    System.out.println ("MvSetModelChecker with lattice:");
    PrintWriter out = new PrintWriter (System.out);    
    getLattice ().dumpValueMap (out);
    out.flush ();

    propOrdering = new AlternatePropOrder (kripke.getPropNames ());
    
    // XXX This should be some how accessed from a generic directory
    // -- Join Irreducible based
    //mvSetFactory = JMvSet.newMvSetFactory (lattice, propOrdering.size ());
    // -- MDD based 
    mvSetFactory = MDDMvSet.newMvSetFactory (lattice, propOrdering.size ());

    mvTransRelation = XKripkeMvSetBuilder.buildMvSet (mvSetFactory, 
						      kripke, 
						      propOrdering);
    cache = new UnboundedMapCache ();

    top = lattice.idtimesIdx ();
    bottom = lattice.idplusIdx ();
    ctlsimplifier = simplifier;
  }

  public MvSetModelChecker (MvSet transitionRelation, 
			    PropOrderIf proporder)
  {
    this (transitionRelation, proporder, new CTLDirectSimplifier ());
  }
  
  public MvSetModelChecker (MvSet transitionRelation,
			    PropOrderIf proporder,
			    CTLSimplifier simplifier)
  {
    lattice = transitionRelation.getLattice ();

    System.out.println ("MvSetModelChecker with lattice:");
    PrintWriter out = new PrintWriter (System.out);    
    getLattice ().dumpValueMap (out);
    out.flush ();

    // set the property ordering
    propOrdering = proporder;

    // set the MvSets
    mvTransRelation = transitionRelation;
    mvSetFactory = mvTransRelation.getFactory ();

    cache = new UnboundedMapCache ();
    top = lattice.idtimesIdx ();
    bottom = lattice.idplusIdx ();
    ctlsimplifier = simplifier;
  }  
    
  public void setTransitionRelation (MvSet v)
  {
    mvTransRelation = v;
  }
  
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public MvSet getMvTransRelation ()
  {
    return mvTransRelation;
  }
  
  /*** This is the top level model checking routine,
   *** eventually no one will use checkCTL any more 
   ***/
  MvSet postStateVars = null;
  
  public MvSet doCTL (CTLNode ctl)
  {
    // -- UGLY as hell... 
    postStateVars = buildPostStateCube (propOrdering.getPrimedMap ());
    postStateVars.resetStats ();
  //    ((MDDMvSet.MDDMvSetFactory)mvSetFactory).getMddManager ().setCaching (true);
    
    MvSet result = checkCTL (ctl);
    postStateVars = null;
    return result;
  }
  
  public MvSet checkCTL (CTLNode ctl)
  {
    MvSet result;
    //    String formula;

    // check formula in cache first
    //    formula = ctl.toFairString();

    if (cache.containsKey (ctl))
      {
	return (MvSet) cache.get (ctl);
      }
    else // formula not cached
      {
	result = computeCTL (ctl);
	cache.put (ctl, result);
	return result;
      }    
  }

  // -- model checking functions
  MvSet computeCTL (CTLNode ctl)
  {

    ctl = simplify (ctl);

    if (ctl instanceof CTLMvSetNode)
      return ((CTLMvSetNode)ctl).getMvSet ();
    else if (ctl instanceof CTLNegNode)      
      return checkNeg (ctl.getRight ());
    else if (ctl instanceof CTLAndNode)
      return checkConjunction (ctl.getLeft (), ctl.getRight ());
    else if (ctl instanceof CTLOrNode)
      return checkDisjunction (ctl.getLeft (), ctl.getRight ());
//      else if (ctl instanceof CTLImplNode)
//        return checkImplies(ctl.getLeft (), ctl.getRight ());
//      else if (ctl instanceof CTLIffNode)
//        return checkEquiv(ctl.getLeft (), ctl.getRight ());
    else if (ctl instanceof CTLEXNode)
      return checkEX (ctl.getRight (), ctl.getFairness ());
    else if (ctl instanceof CTLAXNode)
      return checkAX (ctl.getRight (), ctl.getFairness ());
    else if (ctl instanceof CTLPreEXNode)
      return checkPreEX (ctl.getRight ());
//      else if (ctl instanceof CTLEUiNode)
//        return checkEUi (ctl.getLeft (), ((CTLEUiNode) ctl).getI (),
//  		       ctl.getRight ());
//      else if (ctl instanceof CTLAUiNode)
//        return checkAUi (ctl.getLeft (), ((CTLAUiNode) ctl).getI (),
//  		       ctl.getRight ());
    else if (ctl instanceof CTLEUNode)
      return checkEU (ctl.getLeft (), ctl.getRight (), ctl.getFairness ());
    else if (ctl instanceof CTLAUNode)
      return checkAU (ctl.getLeft (), ctl.getRight (), ctl.getFairness ());
//      else if (ctl instanceof CTLEFNode)
//        return checkEF (ctl.getRight ());
//      else if (ctl instanceof CTLAFNode)
//        return checkAF (ctl.getRight ());
    else if (ctl instanceof CTLEGNode)
      return  checkEG (ctl.getRight (), ctl.getFairness ());
    else if (ctl instanceof CTLAGNode)
      return  checkAG (ctl.getRight (), ctl.getFairness ());
    else if (ctl instanceof CTLEqualsNode)
      return checkEq (ctl.getLeft (), ctl.getRight ());
    else if (ctl instanceof CTLUnderNode)
      return checkUnder (ctl.getLeft (), ctl.getRight ());
    else if (ctl instanceof CTLOverNode)
      return checkOver (ctl.getLeft (), ctl.getRight ());
    else if (ctl instanceof CTLLatticeNode)
      {
	return mvSetFactory.createConstant
	  (((CTLLatticeNode) ctl).getLatIdx ());
      }
    
    else if (ctl instanceof CTLAtomNode)
      {
	CTLAtomNode atom = (CTLAtomNode) ctl; 
	if (propOrdering.getPropOrder (atom.getName ().toString () ) != -1)
	  return checkProp (atom.getName ().toString ());
	else
	  throw new IllegalArgumentException ("Could not understand: " + ctl);
      }
    else
      {
	System.out.println ("Cannot understand: " + ctl);
	System.out.println ("Class type is : " + ctl.getClass ().getName ());
	throw new IllegalArgumentException ("Could not understand");
      }
      
  }  

  // -- needed to simplify CTL into the form that the model checker
  // knows how to deal with
  CTLSimplifier ctlsimplifier;
  CTLNode simplify (CTLNode ctl)
  {
    CTLNode temp = ctl;
    
    // -- expand the CTL until one of the formulas that the
    // -- modelchecker can deal with appears
    do
      {
	ctl = temp;
	temp = ctlsimplifier.simplify (ctl, lattice);
      }
    while (!ctl.equals (temp));
    return ctl;
  }

  CTLNode constructFair (List c)
  {
    CTLNode t = CTLFactory.createCTLLatticeNode (top, lattice);
    CTLNode f = CTLFactory.createCTLLatticeNode (bottom, lattice);

    return CTLFactory.makeSameAndFair (t.eg (), c).under (f).neg ();
  }

  MvSet checkProp (String propName)
  {
    return mvSetFactory.createProjection 
      (propOrdering.getPropOrder (propName));
  }
  MvSet checkEq (CTLNode ctl1, CTLNode ctl2)
  {
    return checkCTL (ctl1).ptwiseCompare (FiniteLattice.EQUAL,
					  checkCTL (ctl2));
  }
  MvSet checkUnder (CTLNode ctl1, CTLNode ctl2)
  {
    return checkCTL (ctl1).ptwiseCompare (FiniteLattice.BELOW,
					  checkCTL (ctl2));
  }  
  MvSet checkOver (CTLNode ctl1, CTLNode ctl2)
  {
    return checkCTL (ctl1).ptwiseCompare (FiniteLattice.ABOVE,
					  checkCTL (ctl2));
  }  

  MvSet checkNeg (CTLNode ctl)
  {
    return checkCTL (ctl).ptwiseNeg ();
  }

  MvSet checkConjunction (CTLNode ctl1, CTLNode ctl2)
  {
    return checkCTL (ctl1).ptwiseCompose (FiniteLattice.MEET, 
					  checkCTL (ctl2));
  }
  
  MvSet checkDisjunction (CTLNode ctl1, CTLNode ctl2)
  {
    return checkCTL (ctl1).ptwiseCompose (FiniteLattice.JOIN, 
					  checkCTL (ctl2));
  }
  
  MvSet checkEX (CTLNode ctl)
  {
    return checkEX (ctl, new LinkedList ());
  }
  MvSet checkEX (CTLNode ctl, List fairness)
  {
    // make some fairness adjustments if necessary
    if (fairness.size () > 0) // there are some fairness conditions
      ctl = ctl.and (constructFair (fairness));
    MvSet Q = checkCTL (CTLFactory.createCTLPreEXNode (ctl));
    return quantify (Q, propOrdering.getPrimedProps ());
  }

  MvSet checkAX (CTLNode ctl)
  {
    return checkAX (ctl, new LinkedList ());
  }
  MvSet checkAX (CTLNode ctl, List fairness)
  {
    // -- computing AX is just as computing EX except we use 
    // -- forall quantification
    MvSet phi = checkCTL (ctl);

    // make some fairness adjustments if necessary
    if (fairness.size () > 0) // there are some fairness conditions
      ctl = ctl.or (constructFair (fairness).neg ());

    MvSet result = 
      mvTransRelation.not ().or 
      (phi.renameArgs (propOrdering.getPrimedMap ()));

    result = forallQuantify (result, propOrdering.getPrimedProps ());
    return result;
  }

  MvSet checkEU (CTLNode ctll, CTLNode ctlr)
  {
    return checkEU (ctll, ctlr, new LinkedList ());
  }
  MvSet checkEU (CTLNode ctll, CTLNode ctlr, List fairness)
  {
    int i = 0;
    MvSet prevEU, curEU;

    // make some fairness adjustments if necessary
    if (fairness.size () > 0) // there are some fairness conditions
      ctlr = ctlr.and (constructFair (fairness));

    // -- initial value of EU
    curEU = checkCTL (ctll.eu (0, ctlr));

    // -- compute the fixpoint
    do
      {
	i++;
	prevEU = curEU;
	curEU = checkCTL (ctll.eu (i, ctlr));
      }
    while (!prevEU.equals (curEU) && i < 1000 /* for safety */);

    return curEU;
  }

  MvSet checkAU (CTLNode ctll, CTLNode ctlr)
  {
    return checkAU (ctll, ctlr, new LinkedList ());
  }
  MvSet checkAU (CTLNode ctll, CTLNode ctlr, List fairness)
  {
    // -- using
    // -- Ac [f U g] = \mu Z. (g \/ !fair) \/ (f /\ AX Z)

    int i = 0;
    MvSet prevAU;
    MvSet curAU;

    // make some fairness adjustments if necessary
    if (fairness.size () > 0) // there are some fairness conditions
      ctlr = ctlr.or (constructFair (fairness).neg ());

    curAU = checkCTL (ctll.au (0, ctlr));
     
    // -- compute fix point
    do 
      {
	i++;
	prevAU = curAU;
	curAU = checkCTL (ctll.au (i, ctlr));
      }
    while (!prevAU.equals (curAU) && i < 1000);
    return curAU;

  }
  
  MvSet original_checkAU (CTLNode ctll, CTLNode ctlr)
  {
    int i = 0;
    MvSet prevAU, curAU;

    // -- initial value of AU
    curAU = checkCTL (ctll.au (0, ctlr));

    // -- compute the fixpoint
    do
      {
	i++;
	prevAU = curAU;
	curAU = checkCTL (ctll.au (i, ctlr));
      }
    while (!prevAU.equals (curAU) & i < 1000 /* for safety */);

    return curAU;
  }

//    MvSet checkEF (CTLNode ctl)
//    {
//      return checkCTL (CTLFactory.createCTLLatticeNode (top, lattice).eu (ctl));
//    }

//    MvSet checkAF (CTLNode ctl)
//    {
//      return checkCTL (CTLFactory.createCTLLatticeNode (top, lattice).au (ctl));
//    }

//    MvSet checkImplies (CTLNode ctl1, CTLNode ctl2)
//    {
//      // a -> b === !a | b
//      return checkCTL (ctl1.neg ().or (ctl2));
//    }
//    MvSet checkEquiv (CTLNode ctl1, CTLNode ctl2)
//    {
//      // a <-> b === a -> b & b -> a
//      return checkCTL ((ctl1.implies (ctl2)).and (ctl2.implies (ctl1)));
//    }

  MvSet checkAG (CTLNode ctl)
  {
    return checkAG (ctl, new LinkedList ());
  }
  MvSet checkAG (CTLNode ctl, List fairness)
  {
    // -- AG f == \nu Z. f /\ AX Z
    // -- AcG f == AG (f \/ !fair)
    MvSet phi = checkCTL (ctl);

    // make some fairness adjustments if necessary
    if (fairness.size () > 0) // there are some fairness conditions
      ctl = ctl.or (constructFair (fairness).neg ());
    
    MvSet curAG = mvSetFactory.createConstant (lattice.idtimesIdx ());
    MvSet prevAG = null;
    
    do {
      prevAG = curAG;
      // -- curAG = f /\ AX prevAG
      MvSet axTerm = 
	forallQuantify (mvTransRelation.not ().or 
			(prevAG.renameArgs (propOrdering.getPrimedMap ())), 
			propOrdering.getPrimedProps ());
      curAG = phi.and (axTerm);
    } while (!curAG.equals (prevAG));
    return prevAG;
  }

  MvSet checkEG (CTLNode ctl)
  {
    return checkEG (ctl, new LinkedList ());
  }


  MvSet oldcheckEG (CTLNode ctl)
  {
    MvSet curEG = mvSetFactory.createConstant (lattice.idtimesIdx ());
    MvSet prevEG = null;

    MvSet f = checkCTL (ctl);
    // XXX Alternatively if our model checker supports MvSet CTL nodes
    // XXX we can rewrite this using checkCTL recursively.
    int count = 0;
    do
      {
	System.out.println ("doing " + ctl + " iteration " + (count++));
	prevEG = curEG;
	// -- curEG = f /\ EX prevEG
	MvSet exTerm = quantify (mvTransRelation.ptwiseCompose 
				 (FiniteLattice.MEET, 
				  prevEG.renameArgs 
				  (propOrdering.getPrimedMap ())), 
				 propOrdering.getPrimedProps ());

	curEG = f.ptwiseCompose (FiniteLattice.MEET, exTerm);
      } while (!curEG.equals (prevEG));
      return prevEG;
  }
  
  MvSet checkEG (CTLNode phi, List fairness)
  {
    MvSet curEG = mvSetFactory.createConstant (lattice.idtimesIdx ());
    MvSet prevEG = null;

    int count = 0;
    do
      {
	log.info ("EGc: doing " + phi + " iteration " + (count++));
	prevEG = curEG;
	// construct the ex terms
	CTLNode z = CTLFactory.createCTLMvSetNode (prevEG);
	CTLNode ctl = phi;
	
	// base case for when there are no fairness conditions
	// perhaps we'll remove this later on.....
	ctl = ctl.and (phi.eu (phi.and (z)).ex ());
	for (Iterator i = fairness.iterator (); i.hasNext ();)
	  {
	    CTLNode c = (CTLNode) i.next ();
	    // .../\ EX E[phi U phi /\ c /\ z]
	    ctl = ctl.and (phi.eu (phi.and (c.and (z))).ex ());
	  }
	curEG = checkCTL (ctl);

      }
    while (!curEG.equals (prevEG));

    return prevEG;
  }
  
  // this is an alternative way of doing EG with fairness
  MvSet checkEGc (CTLNode phi, List fairness)
  {
    MvSet curEG = mvSetFactory.createConstant (lattice.idtimesIdx ());
    MvSet prevEG = null;

    int count = 0;
    do
      {
	log.info ("EGc: doing " + phi + " iteration " + (count++));
	prevEG = curEG;
	// construct the ex terms
	CTLNode z = CTLFactory.createCTLMvSetNode (prevEG);
	CTLNode ctl = z;
	
	// base case for when there are no fairness conditions
	// perhaps we'll remove this later on.....
	ctl = phi.eu (phi.and (ctl)).ex ();

	for (Iterator i = fairness.iterator (); i.hasNext ();)
	  {
	    CTLNode c = (CTLNode) i.next ();
	    // ... EX E[phi U phi /\ c /\ EX...]
	    ctl = phi.eu (phi.and (c.and (ctl))).ex ();
	  }
	ctl = phi.and (ctl);
	curEG = checkCTL (ctl);
      }
    while (!curEG.equals (prevEG));

    return prevEG;
  }

  MvSet original_checkEG (CTLNode ctl)
  {
    // -- use EG f = !AF !f
    return checkCTL (ctl.neg ().af ().neg ());
  }

//    MvSet checkEUi (CTLNode ctll, int i, CTLNode ctlr)
//    {
//      if (i == 0) 
//        {
//  	return checkCTL (ctlr); // return psi
//        }
//      else 
//        {
//  	// -- use E[f Ui p] = p | (f & EX E[f Ui-1 p])
//  	CTLNode result = ctlr.or (ctll.and (ctll. eu (i-1, ctlr).ex ()));
//  	return checkCTL (result);
//        }
//    }

//    MvSet checkAUi (CTLNode ctll, int i, CTLNode ctlr)
//    {
//      if (i == 0) 
//        {
//  	return checkCTL (ctlr); // return psi
//        }
//      else 
//        {
//  	// -- use A[f Ui p] = p | (f & AX A[f Ui-1 p] & EX A[f Ui-1 p])
//  	CTLNode result = ctlr.or (ctll.and 
//  	   (ctll.au (i-1, ctlr).ex ()).and (ctll.au (i-1, ctlr).ax ()));
//  	return checkCTL (result);
//        }
//    }

  MvSet checkPreEX (CTLNode ctl)
    {

      // -- R /\ prime (\phi)
      MvSet result = mvTransRelation.ptwiseCompose 
	(FiniteLattice.MEET, 
	 checkCTL (ctl).renameArgs (propOrdering.getPrimedMap ()));
      return result;
    }
  
  MvSet buildPostStateCube (int[] varIndex)
  {
    MvState state = new MvState (propOrdering.size (), getLattice ());
    for (int i = 0; i < varIndex.length; i++)
      state.setValue (varIndex [i], getLattice ().idtimesIdx ());
    
    return mvSetFactory.createPoint (state, getLattice ().idtimesIdx ());
  }
  
  
  public MvSet forallQuantify (MvSet mvSet, int[] varIndex)
  {
    if (postStateVars != null && varIndex == propOrdering.getPrimedProps ())
      return mvSet.forallAbstract (postStateVars);
    return mvSet.forallAbstract (varIndex);
//    return _forallQuantify (mvSet, varIndex);
  }
  
  public MvSet quantify (MvSet mvSet, int[] varIndex)
  {
    if (postStateVars != null && varIndex == propOrdering.getPrimedProps ())
      return mvSet.existAbstract (postStateVars);
    
    return mvSet.existAbstract (varIndex);
  }
  
  public MvSet _forallQuantify (MvSet mvSet, int[] varIndex)
  {
      for (int i = 0; i < varIndex.length; i++)
	mvSet = forallQuantify (mvSet, varIndex [i]);
      return mvSet;
  }
  
  public MvSet _quantify (MvSet mvSet, int [] varIndex)
    {
      for (int i = 0; i < varIndex.length; i++)
	mvSet = quantify (mvSet, varIndex [i]);
      return mvSet;
    }
 
  public MvSet quantify (MvSet mvSet, int varIndex)
    {
      MvSet result = mvSetFactory.createConstant (lattice.idplusIdx ());
      
      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	result = result.ptwiseCompose 
	  (FiniteLattice.JOIN,
	   mvSet.restrict (varIndex, it.nextInt ()));
      
      return result;
    }

  public MvSet forallQuantify (MvSet mvSet, int varIndex)
  {
    MvSet result = mvSetFactory.createConstant (lattice.idtimesIdx ());
    
    for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
      result = result.and (mvSet.restrict (varIndex, it.nextInt ()));
    return result;
  }
 

  /***
   *** Reduces the transition relation to include only the 
   *** rechable state space
   ***/
  public MvSet computeRechable (MvState initState)
  {
    System.out.println ("ComputeReachable: start");
    /****
     **** This is a fixpoint computation... 
     ****/
    MvSet curQ = initState.toMvSet (mvSetFactory);
    System.out.println ("Created mvset");
    MvSet prevQ;
    
    int count = 1;

    do {
      prevQ = curQ;
      curQ = prevQ.or (forwardImage (mvTransRelation, prevQ));
    } while (!prevQ.equals (curQ));

    /*** curQ contains all of the reachable states 
     *** thus transRel /\ curQ is restriction to reachable states 
     ***/


    return mvTransRelation.and 
      (curQ.eq (mvSetFactory.createConstant (lattice.idplusIdx ())).not ());
  }

  MvSet forwardImage (MvSet trans, MvSet q)
  {
    System.out.println ("Computing forward image");
    MvSet result;
    
    result = trans.and (q);
    System.out.println ("Quantifying the result");
    result = quantify (result, propOrdering.getUnPrimedProps ());
    System.out.println ("Renaming arguments");
    result = result.renameArgs (propOrdering.getUnPrimedMap ());
    System.out.println ("Done with fwd image");
    return result;
//      return quantify (trans.and (q), propOrdering.getUnPrimedProps ()).
//        renameArgs (propOrdering.getUnPrimedMap ());
  }

  /*
  // shouldn't be here probably........
  // list of fairness conditions
  List cf;

  public void setFairness (List f) 
  {
    cf = f;
  }
  */
}


