/***
 *** MvSet implementation based on the CUDD package
 ***/
package edu.toronto.cs.mvset;


import edu.toronto.cs.cudd.*;
import edu.toronto.cs.cudd.Cudd.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;
import java.util.Arrays;

/***
 *** Implementation of an MvSet based on the MDD library
 *** and join irreducible decomposition
 ***/
public class CUBDDMvSet extends AbstractMvSet
{

  static final BooleanLattice boolLattice = new BooleanLattice ();

  CUBDDMvSetFactory factory;
  
  // -- manager for the underlying BDDs
  Cudd cudd;

  // -- Finite lattice corresponding to the manager
  FiniteLattice lattice;
  
  // -- the actuall BDD pointers
  int[] bddPtr;

  // -- join irreducibles of the lattice such that
  // -- mdd [i] represents an l-cut w.r.t. joinIrr [i]
  int[] joinIrr;
  int[] invJoinIrr;
  
  // -- negation table
  int[] neg;
   
  public CUBDDMvSet (CUBDDMvSetFactory _factory, int[] _bddPtr)
  {
    bddPtr = _bddPtr;
    factory = _factory;
    cudd = factory.getCudd ();

    joinIrr = factory.getJoinIrr ();
    invJoinIrr = factory.getInvJoinIrr ();
    neg = factory.getNegTable ();
    lattice = factory.getLattice ();
  }
  

  public void reorder ()
  {
    cudd.reorder (10);
  }

  public int size ()
  {
    return cudd.sharingSize (bddPtr);
  }
  

  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {
    int[] result = new int [joinIrr.length];
    
    int[] gSets = ((CUBDDMvSet)g).getBddPtrs ();

    if (op == FiniteLattice.MEET)
      for (int i = 0; i < result.length; i++)
	result [i] = cudd.bddAndPtr (bddPtr [i], gSets [i]);
    else if (op == FiniteLattice.JOIN)
      for (int i = 0; i < result.length; i++)
	result [i] = cudd.bddOrPtr (bddPtr [i], gSets [i]);
    else
      Assert.assert ("Unknown op: " + op);


    return createMvSet (result);
  }
  
  public MvSet ptwiseCompare (int op, MvSet g)
  { 
    MvSet result = null; // XXX --Victor: this is to make java compiler happy

    switch (op)
      {
      case FiniteLattice.BELOW:
	result = leq (g); break;
      case FiniteLattice.ABOVE:
	result = geq (g); break;
      case FiniteLattice.EQUAL:
	result = eq (g); break;
      default:
	Assert.assert ("Cannot compare using: " + op);
      }
    
    return result;
  }

  public MvSet leq (MvSet g)
  {
    int[] gSets = ((CUBDDMvSet)g).getBddPtrs ();
    
    int bdd = cudd.bddOnePtr ();
    
    for (int i = 0; i < gSets.length; i++)
      {
	int temp2 = cudd.leqPtr (bddPtr [i], gSets [i]);
	int temp1 = cudd.bddAndPtr (bdd, temp2);
	cudd.recursiveDeref (bdd);
	bdd = temp1;
      }
    
    
    int[] result = new int [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      {
	result [i] = bdd;
	cudd.ref (result [i]);
      }
    cudd.recursiveDeref (bdd);
    
    return createMvSet (result);
  }
  
  public MvSet geq (MvSet g)
  {
    int[] gSets = ((CUBDDMvSet)g).getBddPtrs ();
    
    int bdd = cudd.bddOnePtr ();

    for (int i = 0; i < gSets.length; i++)
      {
	int temp2 = cudd.geqPtr (bddPtr [i], gSets [i]);
	int temp1 = cudd.bddAndPtr (bdd, temp2);
	cudd.recursiveDeref (temp2);
	cudd.recursiveDeref (bdd);
	bdd = temp1;
      }
    
    
    int[] result = new int [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      {
	result [i] = bdd;
	cudd.ref (result [i]);
      }
    cudd.recursiveDeref (bdd);

    return createMvSet (result);
  }
  
  public MvSet eq (MvSet g)
  {
    int[] gSets = ((CUBDDMvSet)g).getBddPtrs ();
    
    int bdd = cudd.bddOnePtr ();

    for (int i = 0; i < gSets.length; i++)
      {
	int temp2 = cudd.eqPtr (bddPtr [i], gSets [i]);
	int temp1 = cudd.bddAndPtr (bdd, temp2);
	cudd.recursiveDeref (temp2);
	cudd.recursiveDeref (bdd);
	bdd = temp1;
      }
    
    
    int[] result = new int [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      {
	result [i] = bdd;
	cudd.ref (result [i]);
      }
    cudd.recursiveDeref (bdd);

    return createMvSet (result);
  }
  
  
 
  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    int[] result = new int [joinIrr.length];

    for (int i = 0; i < joinIrr.length; i++)
      result [i] = cudd.bddNotPtr (bddPtr [invJoinIrr [neg [joinIrr [i]]]]);
    return createMvSet (result);
  }
  


  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    int[] result = new int [joinIrr.length];

    //System.arraycopy (bddPtr, 0, result, 0, bddPtr.length);
    for (int i = 0; i < bddPtr.length; i++)
      {
	result [i] = bddPtr [i];
	cudd.ref (result [i]);
      }

    for (int i = 0; i < result.length; i++)
      {
	// -- restrict each argIdx_k variable (k ranges over join irr)
	int startIdx = argIdx * joinIrr.length;
	for (int k = 0; k < joinIrr.length; k++)
	  {
	    int temp  = 
	      cudd.restrictVar (result [i], startIdx + k, 
				lattice.leq (joinIrr [k], value));
	    cudd.recursiveDeref (result [i]);
	    result [i] = temp;
	  }
	
      }

    return createMvSet (result);
  }
  
  public MvSet restrict (MvState state)
  {
    int[] result = new int [joinIrr.length];

    //System.arraycopy (bddPtr, 0, result, 0, bddPtr.length);
    for (int i = 0; i < bddPtr.length; i++)
      {
	result [i] = bddPtr [i];
	cudd.ref (result [i]);
      }
    
    
    // -- for each bdd 
    // -- for each variable
    // -- for each variable_k corresponding to a join irr k
    // -- restrict
    for (int i = 0; i < result.length; i++)
      {
	for (int j = state.getFirstVar (); j <= state.getLastVar (); j++)
	  {
	    if (!state.isSet (j))
	      continue;
	    
	    int startIdx = j * joinIrr.length;
	    for (int k = 0; k < joinIrr.length; k++)
	    {
	      int temp = 
		cudd.restrictVar 
		(result [i], startIdx + k, 
		 lattice.leq (joinIrr [k], state.getValue (j)));
	      
	      cudd.recursiveDeref (result [i]);
	      result [i] = temp;
	    } 
	  }
      }
    

    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] _newArgs)
  {

    int[] newArgs = new int [_newArgs.length * joinIrr.length];

    for (int i = 0; i < _newArgs.length; i++)
      {
	int oldVar = i * joinIrr.length;
	int newVar = _newArgs [i] * joinIrr.length;
	
	for (int k = 0; k < joinIrr.length; k++)
	  newArgs [oldVar + k] = newVar + k;	
      }

    int[] result = new int [joinIrr.length];

    for (int i = 0; i < result.length; i++)
      result [i] = cudd.bddPermutePtr (bddPtr [i], newArgs);
    

    return createMvSet (result);
      //Assert.assert ("Not implemented");
  }
  

  public MvSet existAbstract (MvSet cube)
  {
    return 
      createMvSet (existAbstractPtr (((CUBDDMvSet)cube).getBddPtrs ()[0]));
  }
  
  public MvSet existAbstract (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    
    int cube = cudd.bddOnePtr ();

    for (int i = 0; i < varIndex.length; i++)
      {
	int bddForVar = factory.createBDDForVar (varIndex [i], top);
	int temp = cudd.bddAndPtr (cube, bddForVar);
	
	cudd.recursiveDeref (bddForVar);
	cudd.recursiveDeref (cube);
	cube = temp;
      }
    
    MvSet result = createMvSet (existAbstractPtr (cube));
    cudd.recursiveDeref (cube);
    return result;
  }

  int[] existAbstractPtr (int cube)
  {
    int[] bddPtrs = new int [joinIrr.length];
    
    for (int i = 0; i < bddPtrs.length; i++)
      bddPtrs [i] = cudd.existAbstract (bddPtr [i], cube);
    return bddPtrs;
  }
  
  public MvSet forallAbstract (MvSet cube)
  {
    return 
      createMvSet (forallAbstractPtr (((CUBDDMvSet)cube).getBddPtrs ()[0]));
  }
  
  public MvSet forallAbstract (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    
    int cube = cudd.bddOnePtr ();

    for (int i = 0; i < varIndex.length; i++)
      {
	int bddForVar = factory.createBDDForVar (varIndex [i], top);
	int temp = cudd.bddAndPtr (cube, bddForVar);
	
	cudd.recursiveDeref (bddForVar);
	cudd.recursiveDeref (cube);
	cube = temp;
      }
    
    MvSet result = createMvSet (forallAbstractPtr (cube));
    cudd.recursiveDeref (cube);
    return result;
  }
  int[] forallAbstractPtr (int cube)
  {
    int[] bddPtrs = new int [joinIrr.length];
    
    for (int i = 0; i < bddPtrs.length; i++)
      bddPtrs [i] = cudd.forallAbstract (bddPtr [i], cube);
    return bddPtrs;
  }
  
  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    return evaluate (new MvState (values, lattice));
  }

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);

    int[] bddPtrs = ((CUBDDMvSet)restricted).getBddPtrs ();
    
    int value = lattice.idplusIdx ();
    
    int bddOnePtr = cudd.bddOnePtr ();

    for (int i = 0; i < bddPtrs.length; i++)
      if (bddPtrs [i] == bddOnePtr)
	value = lattice.plus (value, joinIrr [i]);

    cudd.recursiveDeref (bddOnePtr);
    return value;
  }
  


  /*** Arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** Gets the lattice */
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public int[] getBddPtrs ()
  {
    return bddPtr;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (int[] bddPtrs)
  {
    return factory.createMvSet (bddPtrs);
  }
  
  

  public boolean equals (Object o)
  {
    if (! (o instanceof CUBDDMvSet)) return false;
    return equals ((CUBDDMvSet)o);
  }
  
  public boolean equals (CUBDDMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    int[] gSets = v.getBddPtrs ();
    for (int i = 0; i < bddPtr.length; i++)
      if (bddPtr [i] != gSets [i]) return false;
    return true;
  }

  protected void finalize () throws Throwable
  {
    super.finalize ();
    
    //if (2 == 2) return;

    if (cudd.isDead ()) return;

//     System.out.println ("CUBDDMvSet: finalizing");
    for (int i = 0; i < bddPtr.length; i++)
      {
	if (bddPtr [i] != 0)
	  {
// 	    System.out.println ("\tDereferencing: " + bddPtr [i]);
// 	    System.out.println ("\tRefCount: " + cudd.refCount (bddPtr [i]));
	    Assert.assert (cudd.refCount (bddPtr [i]) != 0);
	    //if (cudd.refCount (bddPtr [i] == 0))
	    cudd.recursiveDeref (bddPtr [i]);
// 	    System.out.println ("\tRefCount after: " + 
// 				cudd.refCount (bddPtr [i]));
	    
	    bddPtr [i] = 0;
	  }
      }
//     System.out.println ("CUBDDMvSet: done finalizing");
  }
  
  

  public static MvSetFactory newMvSetFactory (FiniteLattice lattice, 
					      int arrity)
  {
    return new CUBDDMvSetFactory (lattice, arrity);
  }

  public MvSetFactory getFactory ()
  {
    return factory;
  }  

  public static class CUBDDMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    Cudd cudd;
    FiniteLattice lattice;

    int[] joinIrr;
    // -- this is the inverse of joinIrr
    // -- let j be a join irreducible 
    // -- invJoinIrr [j] == i -> joinIrr [i] == j
    // -- it's size is the size of the lattice, 
    // -- and it is not defined for non-join irreducible elements
    int[] invJoinIrr;
    int[] neg;
    
    public CUBDDMvSetFactory (FiniteLattice _lattice,
			      int _arrity)
    {
      arrity = _arrity;
      lattice = _lattice;
      joinIrr = lattice.getJoinIrreducibles ();

      // -- we will have arrity * joinIrr.length * 2 variables
      cudd = new Cudd ((arrity * joinIrr.length) << 1);      

      invJoinIrr = computeInvJoinIrr (lattice.size (), joinIrr);
      
      // -- create the negation table
      neg = new int [lattice.size ()];
      computeNegTable ();

      // -- CUDD expects that all variables are declared prior to use
      // -- we don't want a restriction like that, so here we will create
      // -- all of our variables at once

      // -- an assumption is made that the variables are numbered started
      // -- from 0 in a consequitive fashion
//        for (int i = 0; i < _arrity; i++)
//  	var (i, lattice.idtimesIdx (), lattice.idtimesIdx ());
      
    }
    
    private int[] computeInvJoinIrr (int size, int[] joinIrr)
    {
      int[] invJoinIrr = new int [size];
      for (int i = 0; i < joinIrr.length; i++)
	invJoinIrr [joinIrr [i]] = i;
      return invJoinIrr;
    }
    

    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      int[] bddPtrs = new int [joinIrr.length];

      for (int i = 0; i < joinIrr.length; i++)
	if (lattice.leq (joinIrr [i], value))
	  bddPtrs [i] = cudd.bddOnePtr ();
	else
	  bddPtrs [i] = cudd.bddZeroPtr ();

      return createMvSet (bddPtrs);
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      int[] bddPtrs = new int [joinIrr.length];

      // -- each variable x is now encoded as 
      // -- x_0, x_1, x_2, ... x_n where n is number of join irreducibles
      // -- where x_i is true iff join irreducible i is bellow the value 
      // -- of x
      int startIndex = argIdx * joinIrr.length;
      
      for (int i = 0; i < joinIrr.length; i++)
	bddPtrs [i] = cudd.bddVarPtr (startIndex + i);
      
      return createMvSet (bddPtrs);
    }    
    
    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      int[] bddPtrs = new int [joinIrr.length];

      int bdd = cudd.bddOnePtr ();

      for (int i = 0; i < args.length; i++)
	{
	  if (args [i] == Algebra.NO_VALUE) continue;
	  int bddForVar = createBDDForVar (i, args [i]);
	  int temp = cudd.bddAndPtr (bdd, bddForVar);
	  cudd.recursiveDeref (bdd);
	  cudd.recursiveDeref (bddForVar);
	  bdd = temp;
	}
      

      for (int i = 0; i < bddPtrs.length; i++)
	{
	  if (!lattice.leq (joinIrr [i], value))
	    bddPtrs [i] = cudd.bddZeroPtr ();
	  else
	    {
	      bddPtrs [i] = bdd;
	      cudd.ref (bddPtrs [i]);
	    }
	}
      cudd.recursiveDeref (bdd);

     return createMvSet (bddPtrs);
    }

    // -- create a BDD representation for a variable 
    int createBDDForVar (int var, int val)
    {
      // -- decompose var into joinIrr components, 
      // -- a component is true if the join irreducible is bellow val

      int startIdx = var * joinIrr.length;
      
      int bdd = cudd.bddOnePtr ();
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  int temp;
	  int bddVar = cudd.bddVarPtr (startIdx + i);
	  if (!lattice.leq (joinIrr [i], val))
	    {
	      temp = cudd.bddNotPtr (bddVar);
	      cudd.recursiveDeref (bddVar);
	      bddVar = temp;
	    }
	  temp = cudd.bddAndPtr (bdd, bddVar);
	  cudd.recursiveDeref (bdd);
	  cudd.recursiveDeref (bddVar);
	  bdd = temp;
	}
      return bdd;
    }
    
    public MvSet var (int argIdx, int argVal, int value)
    {
      int bdd = createBDDForVar (argIdx, argVal);
      
      int[] bddPtrs = new int [joinIrr.length];
      
      for (int i = 0; i < bddPtrs.length; i++)
	{
	  if (!lattice.leq (joinIrr [i], value))
	    bddPtrs [i] = cudd.bddZeroPtr ();
	  else
	    {
	      bddPtrs [i] = bdd;
	      cudd.ref (bddPtrs [i]);
	    }
	}
      cudd.recursiveDeref (bdd);
      return createMvSet (bddPtrs);
    }
    

    public MvSet createPoint (MvState state, int value)
    {
      int[] bddPtrs = new int [joinIrr.length];
      
      int bdd = cudd.bddOnePtr ();
      int temp;
      for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
	{
	  if (!state.isSet (i)) continue;
	  int bddForVar = createBDDForVar (i, state.getValue (i));
	  temp = cudd.bddAndPtr (bdd, bddForVar);
	  cudd.recursiveDeref (bdd);
	  cudd.recursiveDeref (bddForVar);
	  bdd = temp;
	}
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  if (!lattice.leq (joinIrr [i], value))
	    bddPtrs [i] = cudd.bddZeroPtr ();
	  else
	    {
	      bddPtrs [i] = bdd;
	      cudd.ref (bddPtrs [i]);
	    }
	}
      cudd.recursiveDeref (bdd);
      return createMvSet (bddPtrs);
    }


    public MvSet createMvSet (int[] bddPtrs)
    {
      return new CUBDDMvSet (this, bddPtrs);
    }

    public FiniteLattice getLattice ()
    {
      return lattice;
    }
    
    public Cudd getCudd ()
    {
      return cudd;
    }
    
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    protected int[] getJoinIrr ()
    {
      return joinIrr;
    }
    protected int[] getInvJoinIrr ()
    {
      return invJoinIrr;
    }
    
    protected int[] getNegTable ()
    {
      return neg;
    }

    private void computeNegTable ()
    {
      int[] f = computeF ();
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  int j = joinIrr [i];
	  neg [lattice.unOp (f [j], FiniteLattice.NEG)] = j;
	}      
    }
    private int[] computeF ()
    {
      int[] f = new int [lattice.size ()];
      
      for (int i = 0; i < joinIrr.length; i++)
	f [joinIrr [i]] = computeF (joinIrr [i]);
      return f;
    }
    private int computeF (int j)
    {
      // -- return max (L - upSet (j))
      int[] upSet = lattice.getUpSet (j);
      
      int result = lattice.idplusIdx ();
      
      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	{
	  int x = it.nextInt ();
	  if (lattice.eq (x, j)) continue;
	  // x < j ==> x \in (L - up (j))
	  else if (lattice.leq (x, j)) 
	    result = lattice.plus (result, x);
	  // x incomparable to j ==> x \in (L - up (j))
	  else if (!lattice.geq (x, j)) 
	    result = lattice.plus (result, x);
	}
      
      return result;
    }
    
    
    
  }
  

}





  
  
