/***
 *** MvSet implementation based on the CUDD package
 *** this is esentially a direct wrapper on ADDs
 *** i.e. only 2 children, many terminal nodes
 ***/
package edu.toronto.cs.mvset;


import edu.toronto.cs.cudd.*;
import edu.toronto.cs.cudd.CuddLatAdd.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.modelchecker.PropOrderIf;


import java.util.Arrays;


/***
 *** Implementation of an MvSet based on the MDD library
 *** and join irreducible decomposition
 ***/
public class CUBiADDMvSet extends AbstractMvSet
{

  CUBiADDMvSetFactory factory;
  
  // -- manager for the underlying BDDs
  CuddLatAdd cudd;

  // -- Finite lattice corresponding to the manager
  Algebra lattice;
  
  // -- the actuall ADD pointer
  int addPtr;

  // -- top and bot elements of the lattice
  int top;
  int bot;
  
   
  public CUBiADDMvSet (CUBiADDMvSetFactory _factory, int _addPtr)
  {
    addPtr = _addPtr;
    factory = _factory;
    cudd = factory.getCudd ();
    lattice = factory.getLattice ();
    top = lattice.idtimesIdx ();
    bot = lattice.idplusIdx ();
  }
  




  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {
    int result = 0;
    
    int gPtr = ((CUBiADDMvSet)g).getAddPtr ();

    if (op == FiniteLattice.MEET)
      result = cudd.addAndPtr (addPtr, gPtr);
    else if (op == FiniteLattice.JOIN)
      result = cudd.addOrPtr (addPtr, gPtr);
    else
      Assert.assert ("Unknown op: " + op);    
    return createMvSet (result);
  }
  
  public MvSet ptwiseCompare (int op, MvSet g)
  { 
    MvSet result = null; // XXX --Victor: this is to make java compiler happy

    switch (op)
      {
      case FiniteLattice.BELOW:
	result = leq (g); break;
      case FiniteLattice.ABOVE:
	result = geq (g); break;
      case FiniteLattice.EQUAL:
	result = eq (g); break;
      default:
	Assert.assert ("Cannot compare using: " + op);
      }
    
    return result;
  }

  public MvSet leq (MvSet g)
  {
    return createMvSet 
      (cudd.addLeqPtr (getAddPtr (), ((CUBiADDMvSet)g).getAddPtr ()));
  }

  public MvSet geq (MvSet g)
  {
    return createMvSet 
      (cudd.addGeqPtr (getAddPtr (), ((CUBiADDMvSet)g).getAddPtr ()));
  }
  
  
  public MvSet eq (MvSet g)
  {
    return createMvSet 
      (cudd.addEqPtr (getAddPtr (), ((CUBiADDMvSet)g).getAddPtr ()));
  }
  

  public int size ()
  {
    return cudd.dagSize (getAddPtr ());
  }
  public void reorder ()
  {
    System.out.println ("Reordering was called, but not yet allowed on ADDs");
    //cudd.reorder (10);
  }
  
 
  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    return createMvSet (cudd.addNotPtr (addPtr));
  }
  


  boolean checkBool (int value)
  {
    if (value == top) return true;
    else if (value == bot) return false;
  
    Assert.assert ("Illegal value: " + value);
    return false;
  }
  
  boolean toBool (int value)
  {
    return value == top;
  }
  
  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    // XXX make sure that reference counting is right
    int result = cudd.restrictVar (addPtr, argIdx, checkBool (value));

    return createMvSet (result);
  }
  
  public MvSet restrict (MvState state)
  {
    int result = addPtr;
    cudd.ref (result);
    

    for (int j = state.getFirstVar (); j <= state.getLastVar (); j++)
      {
	if (!state.isSet (j)) continue;
	
	int val = state.getValue (j);
	int temp = 
	  cudd.restrictVar (result, j, checkBool (state.getValue (j)));

	cudd.recursiveDeref (result);
	result = temp;
      }
    
    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] newArgs)
  {
    return createMvSet (cudd.addPermutePtr (addPtr, newArgs));
  }
  
  

  // XXX forall and exist abstract will not work for ADDs since
  // XXX Cudd implementation uses Times and Plus operations to combine
  // XXX diagrams, and we need meet and join!
  public MvSet existAbstract (MvSet cube)
  {
    System.out.println ("Exist abstract with cube");
    
    MvSet result = 
      createMvSet (cudd.existAbstract (addPtr, 
				       ((CUBiADDMvSet)cube).getAddPtr ()));
    return result;
  }
  
  public MvSet forallAbstract (MvSet cube)
  {
    MvSet result = 
      createMvSet (cudd.forallAbstract (addPtr, 
					((CUBiADDMvSet)cube).getAddPtr ()));
    return result;
  }
  
  
  int buildCube (int[] varIndex)
  {
    int cubePtr = cudd.addOnePtr ();
    
    for (int i = 0; i < varIndex.length; i++)
      {
	int addForVar = 
	  factory.createADDForVar (varIndex [i], top);
	int temp = cudd.addAndPtr (addForVar, cubePtr);
	cudd.recursiveDeref (cubePtr);
	cudd.recursiveDeref (addForVar);
	cubePtr = temp;
      }
    return cubePtr;
  }
  
  public MvSet existAbstract (int[] varIndex)
  {
    System.out.println ("Exist abstract with array");
    int cubePtr = buildCube (varIndex);
    
    int resultPtr = cudd.existAbstract (addPtr, cubePtr);
    cudd.recursiveDeref (cubePtr);
    return createMvSet (resultPtr);
    
  }
  
  public MvSet forallAbstract (int[] varIndex)
  {
    int cubePtr = buildCube (varIndex);
    int resultPtr = cudd.forallAbstract (addPtr, cubePtr);
    cudd.recursiveDeref (cubePtr);
    return createMvSet (resultPtr);
  }
  
  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    return evaluate (new MvState (values, lattice));
  }

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);

    int restrictPtr = ((CUBiADDMvSet)restricted).getAddPtr ();

    return cudd.addGetValue (restrictPtr);
  }
  


  /*** Arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** Gets the lattice */
  public FiniteLattice getLattice ()
  {
    return (FiniteLattice)lattice;
  }

  public int getAddPtr ()
  {
    return addPtr;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (int addPtr)
  {
    return factory.createMvSet (addPtr);
  }
  
  

  public boolean equals (Object o)
  {
    if (! (o instanceof CUBiADDMvSet)) return false;
    return equals ((CUBiADDMvSet)o);
  }
  
  public boolean equals (CUBiADDMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    return addPtr == v.getAddPtr ();
  }


  public String toString ()
  {
    return "CUBiAdd: " + getAddPtr ();
  }
  
  protected void finalize () throws Throwable
  {
    super.finalize ();

    if (addPtr != 0)
      {
	cudd.recursiveDeref (addPtr);
	addPtr = 0;
      }
  }
  

  public void writeToDaVinci (String filename, PropOrderIf propOrder)
  {
    cudd.dumpDaVinciPtr (filename, addPtr);
  }
  

  

  public static MvSetFactory newMvSetFactory (Algebra lattice, int arrity)
  {
    return new CUBiADDMvSetFactory (lattice, arrity);
  }

  public MvSetFactory getFactory ()
  {
    return factory;
  }  

  public static class CUBiADDMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    CuddLatAdd cudd;
    Algebra lattice;

    int top;
    int bot;
    
    
    public CUBiADDMvSetFactory (Algebra _lattice, int _arrity)
    {
      arrity = _arrity;
      lattice = _lattice;

      top = lattice.idtimesIdx ();
      bot = lattice.idplusIdx ();

      cudd = new CuddLatAdd (lattice, arrity);      

    }
    
    

    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      return createMvSet (cudd.addConstant (value));
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      int addPtr = createADDForVar (argIdx, top, top);

      int temp = createADDForVar (argIdx, bot, bot);
      int temp2 = cudd.addOrPtr (addPtr, temp);
      cudd.recursiveDeref (temp);
      cudd.recursiveDeref (addPtr);
      addPtr = temp2;
      return createMvSet (addPtr);
    }    
    
    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      int addPtr = cudd.addConstant (value);
      int temp;
      
      // -- construct the structure
      for (int i = 0; i < args.length; i++)
	{
	  if (args [i] == Algebra.NO_VALUE) continue;
	  
	  int addForVar = createADDForVar (i, args [i]);
	  temp = cudd.addAndPtr (addPtr, addForVar);
	  cudd.recursiveDeref (addPtr);
	  cudd.recursiveDeref (addForVar);
	  addPtr = temp;
	}
     return createMvSet (addPtr);
    }

    public MvSet var (int argIdx, int argVal, int value)
    {
      return createMvSet (createADDForVar (argIdx, argVal, value));
    }
    
    int createADDForVar (int var, int val)
    {
      return createADDForVar (var, val, top);
    }
    
    // -- create a BDD representation for a variable 
    int createADDForVar (int var, int val, int ddVal)
    {
      int addPtr = cudd.addConstant (ddVal);
      int addVar = cudd.addVarPtr (var);
      
      int temp;

      // -- invert val if it is false
      if (val == bot)
	{
	  temp = cudd.addNotPtr (addVar);
	  cudd.recursiveDeref (addVar);
	  addVar = temp;
	}

      temp = cudd.addAndPtr (addVar, addPtr);
      cudd.recursiveDeref (addPtr);
      cudd.recursiveDeref (addVar);
      return temp;
    }
    

    public MvSet createPoint (MvState state, int value)
    {
      int addPtr = cudd.addConstant (value);
      
      int temp;
      for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
	{
	  if (!state.isSet (i)) continue;
	  int addForVar = createADDForVar (i, state.getValue (i));
	  temp = cudd.addAndPtr (addPtr, addForVar);
	  cudd.recursiveDeref (addPtr);
	  cudd.recursiveDeref (addForVar);
	  addPtr = temp;
	}      
      return createMvSet (addPtr);
    }


    public MvSet createMvSet (int addPtr)
    {
      return new CUBiADDMvSet (this, addPtr);
    }

    public FiniteLattice  getLattice ()
    {
      return (FiniteLattice)lattice;
    }
    
    protected CuddLatAdd getCudd ()
    {
      return cudd;
    }
    
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    
  }


}






  
  
