/***
 *** MvSet implementation based on the CUDD package
 ***/
package edu.toronto.cs.mvset;


import edu.toronto.cs.mdd.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;
import java.util.Arrays;

/***
 *** MvSet using ADDs on top of Xbel's DD library
 ***/
public class JADDMvSet extends AbstractMvSet
{

  JADDMvSetFactory factory;
  
  // -- manager for the underlying BDDs
  MDDManager mddManager;

  // -- Finite lattice corresponding to the manager
  FiniteLattice lattice;
  
  // -- the actuall ADD pointer
  MDDNode add;

  int[] joinIrr;
   
  public JADDMvSet (JADDMvSetFactory _factory, MDDNode _add)
  {
    add = _add;
    factory = _factory;
    mddManager = factory.getMddManager ();
    lattice = factory.getLattice ();
    joinIrr = factory.getJoinIrr ();
  }
  




  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {

    MDDNode result = null;
    
    result = mddManager.Apply (add, ((JADDMvSet)g).getAdd (), op);
    return createMvSet (result);
  }
  
  public MvSet ptwiseCompare (int op, MvSet g)
  { 
    MDDNode result = null;
    
    result = mddManager.Apply (add, ((JADDMvSet)g).getAdd (), op);
    return createMvSet (result);
  }

  public MvSet leq (MvSet g)
  {
    return ptwiseCompare (FiniteLattice.BELOW, g);
  }

  public MvSet geq (MvSet g)
  {
    return ptwiseCompare (FiniteLattice.ABOVE, g);

  }
  
  
  public MvSet eq (MvSet g)
  {
    return ptwiseCompare (FiniteLattice.EQUAL, g);
  }
  

  public int size ()
  {
    return mddManager.dagSize (add);
  }
  
 
  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    return createMvSet (mddManager.Apply (add, FiniteLattice.NEG));
  }
  


  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    MDDNode result = add;

    // -- restrict each argIdx_k variable (k ranges over join irr)
    int startIdx = argIdx * joinIrr.length;
    for (int k = 0; k < joinIrr.length; k++)
      {
	int v = lattice.leq (joinIrr [k], value) ? 
	  lattice.idtimesIdx () : lattice.idplusIdx ();
	result = mddManager.Restrict (result, startIdx + k, v);
      }

    return createMvSet (result);
  }
  
  public MvSet restrict (MvState state)
  {
    MDDNode result = add;
    
    
    // -- for each bdd 
    // -- for each variable
    // -- for each variable_k corresponding to a join irr k
    // -- restrict
    for (int j = state.getFirstVar (); j <= state.getLastVar (); j++)
      {
	if (!state.isSet (j))
	  continue;
	    
	int startIdx = j * joinIrr.length;
	for (int k = 0; k < joinIrr.length; k++)
	  {
	    int v = lattice.leq (joinIrr [k], state.getValue (j)) ?
	      lattice.idtimesIdx () : lattice.idplusIdx ();
	    result = mddManager.Restrict (result, startIdx + k, v);
	  } 
      }
    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] _newArgs)
  {
    
    int[] newArgs = new int [_newArgs.length * joinIrr.length];

    for (int i = 0; i < _newArgs.length; i++)
      {
	int oldVar = i * joinIrr.length;
	int newVar = _newArgs [i] * joinIrr.length;
	
	for (int k = 0; k < joinIrr.length; k++)
	  newArgs [oldVar + k] = newVar + k;	
      }

    System.out.println ("Actually renaming arguments");
    return createMvSet (mddManager.renameVars (add, newArgs));
  }
  

  public MvSet existAbstract (MvSet _cube)
  {
    return createMvSet (existAbstract (((JADDMvSet)_cube).getAdd ()));
  }
  
  public MvSet existAbstract (int[] varIndex)
  {
    
    MDDNode cube = buildCube (varIndex);
    return createMvSet (existAbstract (cube));
  }
  
  MDDNode existAbstract (MDDNode cube)
  {
    return mddManager.existAbstract (add, cube);
  }
  

  public MvSet forallAbstract (MvSet _cube)
  {
    return createMvSet (forallAbstract (((JADDMvSet)_cube).getAdd ()));
  }

  MDDNode buildCube (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    
    MDDNode cube = mddManager.getOne ();
    for (int i = 0; i < varIndex.length; i++)
      cube = mddManager.Apply(cube, 
			      factory.createADDForVar (varIndex [i],top, top),
			      FiniteLattice.MEET);
    return cube;
  }
  
  
  public MvSet forallAbstract (int[] varIndex)
  {
    MDDNode cube = buildCube (varIndex);
    return createMvSet (forallAbstract (cube));
  }
  
  MDDNode forallAbstract (MDDNode cube)
  {
    return mddManager.forallAbstract (add, cube);
  }
  


  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    return evaluate (new MvState (values, lattice));
  }

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);

    MDDNode restrictNode = ((JADDMvSet)restricted).getAdd ();

    return restrictNode.getValue ();

  }
  


  /*** Arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** Gets the lattice */
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public MDDNode getAdd ()
  {
    return add;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (MDDNode addNode)
  {
    return factory.createMvSet (addNode);
  }
  
  

  public boolean equals (Object o)
  {
    if (! (o instanceof JADDMvSet)) return false;
    return equals ((JADDMvSet)o);
  }
  
  public boolean equals (JADDMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    return add.equals (v.getAdd ());
  }

  
  

  public static MvSetFactory newMvSetFactory (FiniteLattice lattice, 
					      int arrity)
  {
    return new JADDMvSetFactory (lattice, arrity);
  }

  public MvSetFactory getFactory ()
  {
    return factory;
  }  

  public static class JADDMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    MDDManager mddManager;
    FiniteLattice lattice;

    int[] joinIrr;
    
    public JADDMvSetFactory (FiniteLattice _lattice,
			     int _arrity)
    {
      arrity = _arrity;
      lattice = _lattice;
      joinIrr = lattice.getJoinIrreducibles ();      
      mddManager = new MDDMgr (arrity * joinIrr.length << 1, lattice, 
			       buildIgnoreMap ());
    }
    
    // -- compute the ignore map to ignore everything except
    // -- top and bot
    int[] buildIgnoreMap ()
    {
      if (lattice.size () == 2) return null;
      
      // -- we ignore everything except for top and bottom
      int top = lattice.idtimesIdx ();
      int bot = lattice.idplusIdx ();
      
      int[] map = new int [lattice.size () - 2];
      
      int count = 0;
      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	{
	  int val = it.nextInt ();
	  if (val != top && val != bot)
	    map [count++] = val;
	}
      return map;
      
    }
    

    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      return createMvSet (mddManager.getConstant (value));
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      MDDNode addNode = mddManager.getZero ();
      MDDNode temp;

      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	{
	  int val = it.nextInt ();	  
	  // -- create add for this variable, with value val
	  temp = createADDForVar (argIdx, val, val);
	  
	  // -- OR with our result
	  addNode = mddManager.Apply (addNode, temp, FiniteLattice.JOIN);

	}
      
      return createMvSet (addNode);
    }    
    
    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      MDDNode addNode = mddManager.getConstant (value);
      
      // -- construct the structure
      for (int i = 0; i < args.length; i++)
	{
	  if (args [i] == Algebra.NO_VALUE) continue;
	  MDDNode addForVar = createADDForVar (i, args [i]);
	  addNode = mddManager.Apply (addNode, addForVar, FiniteLattice.MEET);
	}
     return createMvSet (addNode);
    }

    public MvSet var (int argIdx, int argVal, int value)
    {
      return createMvSet (createADDForVar (argIdx, argVal, value));
    }
    
    MDDNode createADDForVar (int var, int val)
    {
      return createADDForVar (var, val, lattice.idtimesIdx ());
    }
    
    // -- create a BDD representation for a variable 
    MDDNode createADDForVar (int var, int val, int ddVal)
    {
      // -- decompose var into joinIrr components, 
      // -- a component is true if the join irreducible is bellow val

      int startIdx = var * joinIrr.length;
      
      MDDNode addNode = mddManager.getConstant (ddVal);
      
      for (int i = 0; i < joinIrr.length; i++)
	{

	  int v = lattice.leq (joinIrr [i], val) ?
	    lattice.idtimesIdx () : lattice.idplusIdx ();
	  
	  MDDNode addVar = mddManager.getVarEquals (startIdx + i, v);
	  addNode = mddManager.Apply (addNode, addVar, FiniteLattice.MEET);
	}
      return addNode;
    }
    

    public MvSet createPoint (MvState state, int value)
    {

      MDDNode addNode = mddManager.getConstant (value);
      
      MDDNode temp;
      for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
	{
	  if (!state.isSet (i)) continue;
	  MDDNode addForVar = createADDForVar (i, state.getValue (i));
	  addNode = mddManager.Apply (addNode, addForVar, FiniteLattice.MEET);
	}
      
      return createMvSet (addNode);
    }


    public MvSet createMvSet (MDDNode addNode)
    {
      return new JADDMvSet (this, addNode);
    }

    public FiniteLattice getLattice ()
    {
      return lattice;
    }
    
    public MDDManager getMddManager ()
    {
      return mddManager;
    }
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    protected int[] getJoinIrr ()
    {
      return joinIrr;
    }  
    
  }
}






  
  
