package edu.toronto.cs.prettygraph;

import java.util.*;

import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.keg.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.mvset.MvState;

// *** NB the factoring/methods used to do the work in this class are a
// tad non-obvious. -- but are what they are for fairly logical
// reasons. Essentially, the visitor either visits a ctl or a state
// node in the KEG graph; it will then call the relevant
// "do{Ctl|State}Node" method.  

// *** the "get{Ctl|State}Node" methods are similar in concept to
// factory methods, providing the objects, AND (And this is why
// they're not factory methods) registering them with the appropriate
// maps, etc., for future reference. Since these maps are used outside
// the factory functions as a reference for the visitor, factoring out to
// a factory class would be tricky as well as useful, I'm afraid.

// *** the "get{}Name" methods are what the "get{}Node" methods farm
// out the task of naming to; they in turn are simply glue to the
// appropriate NamingScheme object.

public class PrettyGraphKegVisitor extends KEGraphVisitorAncestor
  implements KEGraphVisitor
{
  static Logger log = new Logger ("PrettyGraphKegVisitor");

  public static class KegStateNumberingScheme 
    extends NamingScheme.ToupleNumberingScheme
  {
    public KegStateNumberingScheme ()
    {
    }
    
    public Object getName (PrettyKegGraphNode owner, MvState state)
    {
      List tmp = new ArrayList ();
      tmp.add (owner);
      tmp.add (state);
      return super.getName (tmp);
    }
  }

  DGraph daVinci;
  PrettyKegNodeFactory factory;

  // This maps KEG graph nodes to the pretty graph nodes created to
  // represent them (if any).
  Map nodeMap;

  // XXX these should both really be NamingScheme types, I think... -CDTW
  KegStateNumberingScheme statenamingscheme;
  NamingScheme.QuickNumberingScheme ctlnamingscheme;

  public static DGraph getPrettyRep (DGraph graph, PropOrderIf prop, 
				     Map colourMap)
  {
    PrettyGraphKegVisitor v = new PrettyGraphKegVisitor ();
    DGraph temp = v.generate (graph);

    log.debug ("About to colour.");
    PrettyColouringStrategy dvcolour = 
      new PrettyKegColouringStrategy (prop)
      .setEdger (new PrettyKegEdgeColouring(colourMap))
      .setLabeller (new PrettySimpleStateLabelStrategy (prop)
		    .setBanner ("Of Interest:"));
    log.debug ("dvcolour init'd.");
    /*
    for (Iterator it = DGDepthTraverse.getIterator (temp); it.hasNext ();)
      {
	dvcolour.visit ((DGNode)it.next (), null);
	}*/
    //DGHistorianPass.traverseWith (temp, dvcolour);
    DGDoubleDepthTraverse iterativeVisitor = 
      new DGDoubleDepthTraverse (dvcolour, new DGHistorian ());
    iterativeVisitor.traverse (temp);
    return temp;
  }

  //public static DGraph getPrettyRep (DGraph graph, PropOrderIf prop)
  //{
  //  return getPrettyRep (graph, prop, PrettyConstants.samplemap);
  //}

  public PrettyGraphKegVisitor ()
  {
    super ();
    nodeMap = new HashMap ();
    daVinci = new DGraph ( new PrettyKegNodeFactory ());
    factory = (PrettyKegNodeFactory)daVinci.getFactory ();
    statenamingscheme = new KegStateNumberingScheme ();
    ctlnamingscheme = new NamingScheme.QuickNumberingScheme ();
  }

  public DGraph generate (DGraph graph)
  {
    // Iterator firstnodes = graph.getRootNodes ().iterator ();
    log.debug ("PrettyGraph generating on graph" + graph);
    for (Iterator it = DGDepthTraverse.getIterator (graph); it.hasNext (); )
      {
	DGNode node = (DGNode) it.next ();
	log.debug ("generate visit "+node);
	visit (node, null);
      }
    return daVinci;
    
  }

  private void placeNode (DGNode kegnode, PrettyKegGraphNode dvnode)
  {
    nodeMap.put (kegnode, dvnode);
    Assert.assert (kegnode !=null, "kegnode null!");
    Assert.assert (dvnode != null, "dvnode null!");
    log.debug ("Setting Keg for "+dvnode.getName ()+" to "+kegnode);
    dvnode.setKegNode (kegnode);
    log.debug ("   Keg set to "+dvnode.getKegNode ());
    Assert.assert (dvnode.getKegNode ()!= null, "Argl");
  }
  

  protected String getStateName (MvState state, PrettyKegGraphNode owner)
  {
    if (owner == null)
      {
	List l = new ArrayList ();
	l.add (state);
	// XXX Ooh... Subtle to the point of feeling evil... (see below)
	return "S"+statenamingscheme.getName (l);
	// The reason this works: Basically, the way
	// ToupleNumberingScheme is implemented, it doesn't really do
	// any checks on the length of your touple, and will ignore
	// anything further down a touple than the original touple
	// used to set a given element. So, if I getName on a double
	// of (1, 2), and then on a triple of (1, 2, 3), I'll get the
	// double's name. Equally, the system will not object to
	// handling seperatly and individually naming (1, 2) and (a,
	// b, c). Now, since I am consistently using the touple of
	// (state, owner) for identifying "normal" statenodes, if I
	// then use a touple of (owner), I know that they will never
	// overlap (since they're of different types) and moreover
	// there will be no problem with the length of the touple
	// varying. And so, I have a nice way of identifying, and
	// hence labelling, initial nodes. (The assumption being that
	// there will be no two initial nodes with the same
	// state. Which would be meaningless anyways.)
      }
    
    return "S"+statenamingscheme.getName (owner, state);
    // Old stuff. Nicer for debugging.
    /*
    if (owner == null)
      return ("Initial" + PrettyConstants.STATE_CONNECT + state );
    return (owner.getName () + PrettyConstants.STATE_CONNECT + state);
    */
  }
  
  protected String getStateName (KEGStateNode node, PrettyKegGraphNode owner)
  {
    return getStateName (node.getState (), owner);
    
// This stuff is useful for debugging.
    //  if (owner == null)
//      return ("Initial" + PrettyConstants.STATE_CONNECT + node.getState ());
//      return (owner.getName () + PrettyConstants.STATE_CONNECT 
//  	    + node.getState ());
    
  }
  
  protected String getCtlName (KEGCtlNode node, PrettyKegGraphNode owner)
  {
    return "C"+ctlnamingscheme.getName (node);
    
      // This stuff is useful for debugging...
   //   if (owner == null)
//        return ("CTLInitial" + PrettyConstants.CTL_CONNECT + node.getCTL ());
//      return (owner.getName () + PrettyConstants.CTL_CONNECT + node.getCTL ());
    
  }

  protected void doCtlNode (KEGCtlNode kegnode, PrettyKegGraphNode supernode)
  {
    PrettyKegGraphNode newnode = getDVNode (kegnode, supernode);
    supernode.add(newnode, new PrettyGraphEdge (supernode, newnode));
  }

  protected void doCtlNode (KEGCtlNode kegnode)
  {
    PrettyKegGraphNode newnode = (PrettyKegGraphNode) 
      ((PrettyKegGraphNode) factory.createNode (getCtlName (kegnode, null)))
      .setOwner (null)
      .setType (PrettyConstants.CTL_TYPE);
    daVinci.addRootNode (newnode);
  } 
  
  protected void doStateNode (KEGStateNode kegnode, PrettyKegGraphNode dvnode)
  {
    PrettyKegGraphNode succnode = getDVNode (kegnode, dvnode); // Method 2
      //(PrettyKegGraphNode)
      //factory.createNode (getStateName (kegnode, dvnode))
      //.setType (PrettyConstants.ST_TYPE);    
    dvnode.add (succnode, new PrettyKegEdge (dvnode, succnode, 
					     (dvnode.getKegNode ())
					     .getEdge (kegnode) ));
    //placeNode (kegnode, succnode);
  }

  protected void doStateNode (KEGStateNode kegnode)
  {
    PrettyKegGraphNode succnode = (PrettyKegGraphNode)
      ((PrettyKegGraphNode) factory.createNode (getStateName (kegnode, null)))
        .setType (PrettyConstants.ST_TYPE);
    succnode.setOwner (succnode);
    daVinci.addRootNode (succnode);
    placeNode (kegnode, succnode);
  }

  protected void doStateNode (MvState state, PrettyKegGraphNode supernode)
  {
    PrettyKegGraphNode newnode = getDVNode (state, supernode);
    // System.err.println ("Hi, we do actually add the edge..."
    //		+ supernode + " to " + newnode);	
    supernode.add 
      (newnode, new PrettyKegEdge
	(supernode, newnode, 
	 newnode.getOwner ().getEdge (newnode.getKegNode ())) );
  }
  

  private PrettyKegGraphNode getDVNode (KEGCtlNode kegnode, 
					PrettyKegGraphNode sn)
  {
    PrettyKegGraphNode dvowner = sn.getOwner ();
    PrettyKegGraphNode newnode = (PrettyKegGraphNode) 
      ((PrettyKegGraphNode) factory.createNode 
      (getCtlName (kegnode, dvowner)))
      .setOwner (dvowner)
      .setKegNode (kegnode)
      .setType (PrettyConstants.CTL_TYPE);
    log.debug ("DVG: Created c-node with name: "+newnode.getName ());
    placeNode (kegnode, newnode);
    return newnode;
  }
  
  private PrettyKegGraphNode getDVNode (MvState kegstate, 
					PrettyKegGraphNode sn)
  {
    PrettyKegGraphNode owner = sn.getOwner ();
    PrettyKegGraphNode newnode = (PrettyKegGraphNode)
      ((PrettyKegGraphNode) factory.createNode
       (getStateName (kegstate, owner)))
      .setType (PrettyConstants.ST_TYPE); // NNN Why do I do this here?
    newnode = newnode.setOwner (newnode);
    // placeNode (kegnode, succnode);
    log.debug ("DVG: Created s-node with name: "+newnode.getName ());
    return newnode;
  }

  private PrettyKegGraphNode getDVNode (KEGStateNode kegstate, 
					PrettyKegGraphNode sn)
  {
    PrettyKegGraphNode owner = sn.getOwner ();
    PrettyKegGraphNode newnode = (PrettyKegGraphNode)
      ((PrettyKegGraphNode) factory.createNode
       (getStateName (kegstate, owner)))
      .setType (PrettyConstants.ST_TYPE); // NNN Why do I do this here, and is it right?
    newnode = newnode.setOwner (newnode)
      .setKegNode (kegstate);
    placeNode (kegstate, newnode);
    log.debug ("DVG: Created s-node with name -METHOD 2- : "+newnode.getName ());
    return newnode;
  }
  
  
  public void visitKCtlNode (KEGCtlNode node, Object stateinfo)
  {
    log.debug ("visit ctl");
    if ((PrettyKegGraphNode)nodeMap.get (node) == null)
      doCtlNode (node);

    for (Iterator it = node.getSuccessors ().iterator ();
	 it.hasNext ();)
      {
	PrettyKegGraphNode supernode = (PrettyKegGraphNode)nodeMap.get (node);
	KEGCtlNode kegnode = (KEGCtlNode)it.next ();
	doCtlNode (kegnode, supernode);
	// PrettyKegGraphNode newnode = createDVNode (kegnode, supernode);
	// supernode.add(newnode, new PrettyGraphEdge (supernode, newnode));
	// visit (kegnode);
	// dvcolour.colour (newnode);
      }
    for (Iterator it = node.getStates ().iterator ();
	 it.hasNext ();)
      {
        PrettyKegGraphNode supernode = (PrettyKegGraphNode)nodeMap.get (node);
	MvState state = (MvState)it.next ();
	doStateNode (state, supernode);
      }
    
  }

  public void visitKStateNode (KEGStateNode node, Object stateinfo)
  {
    log.debug ("visit state "+node);
    if (node.isRootNode ())
      doStateNode (node);
    
    for (Iterator it = node.getCTLGraph ().getRootNodes ().iterator ();
	 it.hasNext ();)
      {
	KEGCtlNode kegnode = (KEGCtlNode)it.next ();
	PrettyKegGraphNode supernode = (PrettyKegGraphNode)nodeMap.get (node);
	log.debug ("supernode = "+ supernode.getName ());
	doCtlNode (kegnode, supernode);
      }
    for (Iterator it = node.getSuccessors ().iterator ();
	 it.hasNext ();)
      {
	PrettyKegGraphNode supernode = (PrettyKegGraphNode)nodeMap.get (node);
	KEGStateNode kegnode = (KEGStateNode)it.next ();
	doStateNode (kegnode, supernode);
      }
    generate (node.getCTLGraph ());
  }

  public void visitDGNode (DGNode node, Object stateinfo)
  {
    // Ok, this sort of makes the whole business of being a visitor
    // seem a tad redundant, but there you go. It still makes sense! :-)
    Assert.assert (false, 
		   "DVGraphVisitor: the graph passed was not a keg graph!");
  }
  
}


















